"""Continuous sources"""

from ss2hcsp.sl.sl_block import SL_Block
from ss2hcsp.hcsp.expr import AVar, AConst, FunExpr


class Sine(SL_Block):
    """Sine wave"""
    def __init__(self, name, amplitude, bias, frequency, phase, st):
        super(Sine, self).__init__("sine", name, 1, 0, st)

        self.amplitude = amplitude
        self.bias = bias
        self.frequency = frequency
        self.phase = phase

    def get_expr(self):
        """Obtain the expression corresponding to the block."""
        e = AVar("t")
        if self.frequency != 1:
            e = AConst(self.frequency) * e
        if self.phase != 0:
            e = e + AConst(self.phase)
        e = FunExpr("sin", [e])
        if self.amplitude != 1:
            e = AConst(self.amplitude) * e
        if self.bias != 0:
            e = e + AConst(self.bias)
        return e

    def __str__(self):
        out_var = self.src_lines[0][0].name
        expr = self.get_expr()
        return "%s: %s = %s  (st = %s)" % (self.name, out_var, expr, self.st)

    def __repr__(self):
        return "Sine(%s, %s, %s, %s, %s, %s, out = %s)" % \
            (self.name, self.amplitude, self.bias, self.frequency, self.phase, self.st,
             str(self.src_lines))

    def get_var_subst(self):
        out_var = self.src_lines[0][0].name
        expr = self.get_expr()
        return {out_var: expr}



"""Unit test for translation to Isabelle."""

import unittest

from ss2hcsp.hcsp import isabelle
from ss2hcsp.hcsp import parser
from ss2hcsp.hcsp import pprint


class IsabelleTest(unittest.TestCase):
    def testTranslate(self):
        with open('CaseStudies/lander/lander_simple.txt', 'r', encoding='utf-8') as f:
            text = f.read()
        process = parser.parse_file(text)
        output = isabelle.translate_isabelle(process, "Lander1")
        with open('CaseStudies/lander/Lander1.thy', 'w', encoding='utf-8') as f:
            f.write(output)

    def testTranslate2(self):
        with open('CaseStudies/lander/lander_simple2.txt', 'r', encoding='utf-8') as f:
            text = f.read()
        process = parser.parse_file(text)
        output = isabelle.translate_isabelle(process, "Lander2")
        with open('CaseStudies/lander/Lander2.thy', 'w', encoding='utf-8') as f:
            f.write(output)

    def testTranslate3(self):
        with open('CaseStudies/lander/lander_simple3.txt', 'r', encoding='utf-8') as f:
            text = f.read()
        process = parser.parse_file(text)
        output = isabelle.translate_isabelle(process, "Lander3")
        with open('CaseStudies/lander/Lander3.thy', 'w', encoding='utf-8') as f:
            f.write(output)

    def testTranslate4(self):
        with open('CaseStudies/lander/lander_simple4.txt', 'r', encoding='utf-8') as f:
            text = f.read()
        process = parser.parse_file(text)
        output = isabelle.translate_isabelle(process, "Lander4")
        with open('CaseStudies/lander/Lander4.thy', 'w', encoding='utf-8') as f:
            f.write(output)


if __name__ == "__main__":
    unittest.main()



import unittest
from ss2hcsp.hcsp import hcsp
from ss2hcsp.hcsp.parser import hp_parser
import re


class SystemCTest(unittest.TestCase):
    def testSystemC(self):
        def systemCform(_hp, _name=""):
            if isinstance(_hp, hcsp.ITE):
                head_cond, head_hp = _hp.if_hps[0]

                if isinstance(head_hp, hcsp.Var):
                    new_name_0 = head_hp.name
                else:
                    new_name_0 = "SC" + str(len(indices))
                    indices.append(new_name_0)
                    _ = systemCform(head_hp, new_name_0)

                if len(_hp.if_hps) >= 2:
                    new_name_1 = "SC" + str(len(indices))
                    indices.append(new_name_1)
                    _ = systemCform(hcsp.ITE(_hp.if_hps[1:], _hp.else_hp), new_name_1)
                else:
                    assert len(_hp.if_hps) == 1
                    if isinstance(_hp.else_hp, hcsp.Var):
                        new_name_1 = _hp.else_hp.name
                    elif _hp.else_hp is not None:
                        new_name_1 = "SC" + str(len(indices))
                        indices.append(new_name_1)
                        _ = systemCform(_hp.else_hp, new_name_1)

                _hp.if_hps = ((head_cond, hcsp.Var(new_name_0)),)
                if _hp.else_hp is not None:
                    _hp.else_hp = hcsp.Var(new_name_1)

            elif isinstance(_hp, hcsp.ODE_Comm):
                io_comms = list()
                for ch, sub_hp in _hp.io_comms:
                    if isinstance(sub_hp, hcsp.Var):
                        io_comms.append((ch, sub_hp))
                    else:
                        new_name = "SC" + str(len(indices))
                        indices.append(new_name)
                        _ = systemCform(sub_hp, new_name)
                        io_comms.append((ch, hcsp.Var(new_name)))

                _hp.io_comms = io_comms

            elif isinstance(_hp, hcsp.Sequence):
                hps = list()
                for sub_hp in _hp.hps:
                    hps.append(systemCform(sub_hp))
                _hp.hps = hps

            elif isinstance(_hp, hcsp.Loop):
                if not isinstance(_hp.hp, hcsp.Var):
                    new_name = "SC" + str(len(indices))
                    indices.append(new_name)
                    _ = systemCform(_hp.hp, new_name)
                    _hp.hp = hcsp.Var(new_name)

            elif isinstance(_hp, hcsp.SelectComm):
                io_comms = list()
                for ch, sub_hp in _hp.io_comms:
                    if isinstance(sub_hp, (hcsp.Var, hcsp.Skip)):
                        io_comms.append((ch, sub_hp))
                    else:
                        io_comms.append((ch, systemCform(sub_hp)))
                _hp.io_comms = io_comms

            sc_process.add(_name, _hp)
            return _hp

        expected_process = hcsp.HCSPProcess()
        f = open("./Examples/AADL/isolette/new_hcsp4.txt", "r")
        for line in f.readlines():
            start, end = re.search(pattern="::=", string=line).span()
            name = line[:start].strip()
            process = line[end:].strip()
            expected_process.add(name, hp_parser.parse(process))
        f.close()

        sc_process = hcsp.HCSPProcess()
        indices = []
        for name, process in expected_process.hps:
            _ = systemCform(process, name)
        new_hps = [(name, process) for name, process in sc_process.hps if name]
        sc_process.hps = new_hps
        print(sc_process.sc_str())

        f = open("./Examples/AADL/isolette/isolette.hcsp", "w")
        f.write(sc_process.sc_str())
        f.close()


if __name__ == "__main__":
    unittest.main()



from xml.dom.minidom import parse as XMLparse
import codecs
import json
import tempfile


def parse_aadlmodel(model):
    """Return instances of the model separated by type."""
    nodes = model.childNodes
    features = [node for node in nodes if node.nodeName == 'featureInstance']
    components = [node for node in nodes if node.nodeName == 'componentInstance']
    connections = [node for node in nodes if node.nodeName == 'connectionInstance']
    opas = [node for node in nodes if node.nodeName == 'ownedPropertyAssociation']

    return features, components, connections, opas

def getFeatures(features):
    """Interpret a list of features."""
    Feas = []
    for feature in features:
        fea = {}
        fea['name'] = feature.getAttribute('name')
        if feature.getAttribute('direction') != "":
            fea['direction'] = feature.getAttribute('direction')
        else:
            fea['direction'] = 'in'
        fea['type'] = feature.getElementsByTagName('feature')[0].getAttribute('xsi:type').split(':')[-1]
        Feas.append(fea)
    return Feas

def aadl_parser(aadl_string):
    """Parse the given XML file, add the read info into dic."""
    with tempfile.NamedTemporaryFile("w+") as in_f:
        in_f.write(aadl_string)
        in_f.flush()
        in_f.seek(0)
        parse_model = XMLparse(in_f)

    out_file = dict()
    model = parse_model.getElementsByTagName('instance:SystemInstance')[0]
    category = model.getAttribute('category')
    model_name = model.getAttribute('name')
    out_file["model_name"] = category + "_" + model_name

    features, components, connections, opas = parse_aadlmodel(model)
    Feas = getFeatures(features)

    print(Feas)
    inport_names = []
    outport_names = []
    for fea in Feas:
        if fea["type"] == "DataPort":
            if fea["direction"] == "in":
                inport_names.append(fea["name"])
            if fea["direction"] == "out":
                outport_names.append(fea["name"])
        else:
            continue
    out_file["Inport"] = inport_names
    out_file["Outport"] = outport_names

    return out_file


if __name__=="__main__":

    aadl_parser('aadl_isolette.xml', 'aadl_isolette.json')





"""
Python implementation of Hybrid Hoare Logic.

This version uses data structures that are separate from the main
HCSP program.

"""
import copy

from sympy import sympify, degree, symbols, factor_list, fraction, simplify

from hhlpy.sympy_wrapper import sp_polynomial_div, sp_simplify, sp_is_polynomial
from hhlpy.wolframengine_wrapper import solveODE, wl_prove
from hhlpy.wolframengine_wrapper import wl_simplify, wl_polynomial_div, wl_is_polynomial, found_wolfram
from hhlpy.z3wrapper import z3_prove
from ss2hcsp.hcsp import hcsp, expr, assertion, label
from ss2hcsp.hcsp.parser import expr_parser, expr_parser
from ss2hcsp.hcsp.simulator import get_pos


def compute_diff(e, eqs_dict, functions):
    """Compute differential of an arithmetic or boolean expression."""
    def rec(e):
        if isinstance(e, expr.LogicExpr):
            if e.op == "&&":
                return expr.LogicExpr("&&", rec(e.exprs[0]), rec(e.exprs[1]))
            elif e.op == "||":
                return expr.LogicExpr("&&", rec(e.exprs[0]), rec(e.exprs[1]))
            elif e.op == "->":
                return rec(expr.LogicExpr("||", expr.neg_expr(e.exprs[0]), e.exprs[1]))
            elif e.op == "!":
                return rec(expr.neg_expr(e.exprs[0]))
            else:
                raise NotImplementedError
        elif isinstance(e, expr.RelExpr):
            if e.op == '<' or e.op == '<=':
                return expr.RelExpr("<=", rec(e.expr1), rec(e.expr2))
            elif e.op == '>' or e.op == '>=':
                return expr.RelExpr(">=", rec(e.expr1), rec(e.expr2))
            elif e.op == '==' or e.op == '!=':
                return expr.RelExpr("==", rec(e.expr1), rec(e.expr2))
            else:
                raise NotImplementedError
        elif isinstance(e, expr.AConst):
            return expr.AConst(0)
        elif isinstance(e, expr.BConst):
            return expr.BConst(True)
        elif isinstance(e, expr.FunExpr):
            if len(e.exprs) == 0:
                return expr.AConst(0)      
            elif e.fun_name in functions:
                return rec(expr.replace_function(e, functions))
            else:
                raise NotImplementedError
        elif isinstance(e, expr.AVar):
            if e.name in eqs_dict:
                return eqs_dict[e.name]
            else:
                return expr.AConst(0)
        elif isinstance(e, expr.OpExpr):
            if len(e.exprs) == 1:
                return expr.OpExpr("-", rec(e.exprs[0]))
            elif e.op == '+':
                return expr.OpExpr("+", rec(e.exprs[0]), rec(e.exprs[1]))
            elif e.op == '-':
                return expr.OpExpr("-", rec(e.exprs[0]), rec(e.exprs[1]))
            elif e.op == '*':
                # d(u*v) = u*dv + du*v
                du, dv = rec(e.exprs[0]), rec(e.exprs[1])
                return expr.OpExpr("+", expr.OpExpr("*", e.exprs[0], dv), 
                                        expr.OpExpr("*", du, e.exprs[1]))
            elif e.op == '^' and isinstance(e.exprs[1], expr.AConst):
                # d(u^n) = n * u^(n-1) * du
                du = rec(e.exprs[0])
                mid_expr = expr.OpExpr("^", e.exprs[0], expr.OpExpr("-", e.exprs[1], expr.AConst(1)))
                return expr.OpExpr("*", e.exprs[1], expr.OpExpr("*", mid_expr, du)) 
            elif e.op == '/':
                # d(u/v) = (u*dv + du * v)/ v^2
                # If v is constant, d(u/v) = 1/v * du, which is easier than using above rule.
                if isinstance(e.exprs[1], expr.AConst) or \
                   isinstance(e.exprs[1], expr.FunExpr) and len(e.exprs == 0): # actually a constant
                    du = rec(e.exprs[0])
                    return expr.OpExpr('*', expr.OpExpr('/', expr.AConst(1), e.exprs[1]), du)
                else:
                    du = rec(e.exprs[0])
                    dv = rec(e.exprs[1])
                    numerator = expr.OpExpr("+", expr.OpExpr("*", e.exprs[0], dv), expr.OpExpr("*", du, e.exprs[1]))
                    denominator = expr.OpExpr('*', e.exprs[1], e.exprs[1])
                    return expr.OpExpr('/', numerator, denominator)            
            else:
                raise NotImplementedError
        else:
            raise NotImplementedError
    return rec(e)

def constraint_examination(e):
    '''Examine whether the constraint is open intervals or not.'''
    if not isinstance(e, expr.Expr):
        raise NotImplementedError
    
    def rec(e):
        if isinstance(e, expr.RelExpr):
            if e.op in [">", "<", "!="]:
                return True
            else:
                return False
        elif isinstance(e, expr.LogicExpr):
            if e.op == '!':
                return not rec(e.exprs[0])
            elif e.op == '&&' or e.op == '||':
                return rec(e.exprs[0] and e.exprs[1])
    return rec(e)

def compute_boundary(e):
    # Compute the boundary for a boolean expression.
    if isinstance(e, expr.RelExpr):
        if e.op in ['<', '>', '!=']:
            return expr.RelExpr("==", e.expr1, e.expr2)
    elif isinstance(e, expr.LogicExpr):
        if e.op == '&&':
            boundary1 = compute_boundary(e.exprs[0])
            boundary2 = compute_boundary(e.exprs[1])
            disj1 = expr.LogicExpr('&&', e.exprs[0], boundary2)
            disj2 = expr.LogicExpr('&&', e.exprs[1], boundary1)
            disj3 = expr.LogicExpr('&&', boundary1, boundary2)
            return expr.list_disj(disj1, disj2, disj3)
        elif e.op == '||':
            boundary1 = compute_boundary(e.exprs[0])
            boundary2 = compute_boundary(e.exprs[1])
            neg1 = expr.neg_expr(e.exprs[0])
            neg2 = expr.neg_expr(e.exprs[1])
            disj1 = expr.LogicExpr('&&', neg1, boundary2)
            disj2 = expr.LogicExpr('&&', neg2, boundary1)
            return expr.LogicExpr('||', disj1, disj2)
        elif e.op == '!':
            return compute_boundary(expr.neg_expr(e.exprs[0]))

# Return the relexpression 'denomibator != 0' for term e.           
def demoninator_not_zero(e):
    if not isinstance(e, expr.Expr):
        raise NotImplementedError

    if not isinstance(e, str):
        e = str(e)
    
    e = simplify(sympify(e.replace('^', '**')))
    _, denominator = fraction(e)
    denominator = expr_parser.parse(str(denominator).replace('**', '^'))

    return expr.RelExpr('!=', denominator, expr.AConst(0))

# Create a new AVar by adding suffix.
# s is the root var name, names is the set of names being used.
def create_newvar(s, names):
    if not isinstance(s, str):
        raise AssertionError
    # If s is already a new variable name, return the AVar object of s.
    if s not in names:
        return expr.AVar(s)
    # Increase the suffix number until it's a new variable name.
    suffix_n = 0
    while (s + str(suffix_n) in names):
        suffix_n = suffix_n + 1
    return expr.AVar(s + str(suffix_n))


class CmdInfo:
    """Information associated to each HCSP program."""
    def __init__(self):
        # Pre-condition and post-condition are assertions.
        self.pre = None
        self.post = None

        # List of verification conditions for this command.
        self.vcs = []

        # Type of the parent program.
        self.parent_type = None

        # Equation dict for ODEs.
        self.eqs_dict = dict()

        # Assumptions for HCSPs.
        self.assume = []

        # Invariants for loop at this position.
        self.inv = []

        # Use differential weakening rule or not
        # It's True by default, which means we apply dw rule on ODE by default.
        # After applying dw rule, set dw to be False and one of the other rules to be True to verify invariants. 
        self.dw = True

        # Use solution axiom or not
        self.sln_rule = False

        # Use the trivial semantics or not.(True and False is the invariant for all ODEs)
        self.tv = False

        # Use differiential invariant rule.
        self.dI_rule = False

        # Invariants in differiential invariant for ODEs.
        self.dI_inv = None

        # Differential cuts for ODEs.
        self.diff_cuts = []

        # dG invariants for ODEs.
        self.dG_inv = None

        # Ghost invariants for new ODEs in which ghost equation is added.
        self.ghost_inv = None

        # Name of the ghost variable
        self.ghost_var = None

        # Computed differential equations for ghosts.
        self.ghost_eqs = dict()

        # Use darboux rule for ODEs.
        self.dbx_rule = False

        # Invariants in dbx rule or ODEs.
        self.dbx_inv = None

        # Dbx cofactor for ODEs
        self.dbx_cofactor = None

        # Use barrier certificate rule or not
        self.barrier_rule = False 

        # Add barrier invariant for ODEs.
        self.barrier_inv = None

        # Use conjunction rule
        self.conj_rule = False

        # Use andR rule
        self.andR = False

    def __str__(self):
        res = ""
        res += "pre = %s\n" % self.pre
        res += "post = %s\n" % self.post
        for vc in self.vcs:
            res += "vc: %s\n" % vc
        if self.inv is not None:
            res += "inv: %s\n" % self.inv
        if self.ghost_inv is not None:
            res += "ghost_inv: %s\n" % self.ghost_inv
        if self.ghost_eqs is not None:
            for ghost_var, eq in self.ghost_eqs.items():
                res += "ghost_eq: %s_dot = %s\n" % (ghost_var, str(eq))
        for diff_cut in self.diff_cuts:
            res += "diff_cut: %s\n" % diff_cut

        return res

class Condition:
    """Stores a condition(pre-condition, post-condition, verification condition) and the parts of the program that it originates from."""
    def __init__(self, expr, path=[], origins=[], blabel=None, post_blabel=None, bottom_loc=None, categ=None, isVC=False):
        # The expression stating the condition itself
        self.expr = expr
        # The positions of hps where the condition originates from
        self.path = path
        # The locations of assertions and expressions that the condition originates from
        self.origins = origins
        # The location of the bottom-most assertion assocated with the condition
        self.bottom_loc = bottom_loc
        # The category of VC, used in VC generated by loop and ODE,
        # which can be "init" or "maintain"
        self.categ = categ

        # The branch label of the last post condition from which the wp is computed from,
        # only acquired when the Condition is a computed wp, otherwise, set as None by default.
        self.post_blabel = post_blabel

        # The branch label of this condition 
        self.blabel = blabel

        # Whether this condition is a verification condition or not
        self.isVC = isVC

        if self.isVC:
            # Compute the composite label when vc is True
            if self.categ and self.blabel:
                self.comp_label = label.CompLabel(categ_label=label.CategLabel(categ=self.categ),
                                                  branch_label=self.blabel)
            elif self.categ and not self.blabel:
                self.comp_label = label.CategLabel(categ=self.categ)
            elif not self.categ and self.blabel:
                self.comp_label = self.blabel
            else:
                self.comp_label = None

            # The solver used to verify this verification condition.
            # Set as 'z3' by default.
            self.solver = 'z3'

    def __str__(self):
        return str(self.expr)


class OriginLoc:
    """Location of an assertion(pre-condition, post-condition, invariant) or expression
    that a Condition originates from"""
    def __init__(self, index=None, isPre=False, isPost=False, isInv=False, isConstraint=False, hp_pos=None):
        # The index of the assertion
        self.index = index
        # Whether it is pre-condition of the whole program
        self.isPre = isPre
        # Whether it is post-condition of the whole program
        self.isPost = isPost
        # Whether it is invariant
        self.isInv = isInv
        # Whether it is ode constraint
        self.isConstraint = isConstraint
        # Only required if it is an invariant or constraint.
        # The position of the hcsp program, to which the invariant is attached.
        self.hp_pos = hp_pos


class CmdVerifier:
    """Contains current state of verification of an HCSP program."""
    def __init__(self, *, pre, hp, post, functions=dict()):
        # The HCSP program to be verified.
        self.hp = hp

        # The list of post conditions of the whole HCSP program
        self.post = post

        # Mapping from program position to CmdInfo objects.
        self.infos = dict()

        # The conjunction of post expression
        pre_conj = expr.list_conj(*[subpre.expr for subpre in pre])
        # The conjunction of post expression
        post_conj = expr.list_conj(*[subpost.expr for subpost in post])

        # Dictionary of declared functions
        # TODO: Use this below instead of scanning the expressions?
        self.functions = functions

        # Set of function names that are used
        fun_names = hp.get_fun_names().union(pre_conj.get_fun_names(), post_conj.get_fun_names())

        # Set of var_names that are used, the names of AVar expression 
        var_names = hp.get_vars().union(pre_conj.get_vars(), post_conj.get_vars())

        # Set of names(function names and var_names) that are used
        self.names = fun_names.union(var_names)

        # Set of program variables that are used
        self.variable_names = set()
        self.compute_variable_set()

        # Set of functions with zero arity that are used, which can be treated as constants
        zero_arity_funcs = hp.get_zero_arity_funcs().union(pre_conj.get_zero_arity_funcs(), post_conj.get_zero_arity_funcs())

        self.constant_names = zero_arity_funcs.union(var_names - self.variable_names)

        # Initialize info for the root object. Place pre-condition and post-condition.
        # pos is a pair of two tuples. 
        # pos[0] is the position of the sub_hp in hp.
        # pos[1] is the number of subproof of the same sub_hp
        root_pos = ((),())
        self.infos[root_pos] = CmdInfo()
        self.infos[root_pos].pre = [Condition(expr=subpre.expr, 
                                              origins=[OriginLoc(isPre=True, index=idx)]) 
                                        for idx, subpre in enumerate(pre)]
        self.infos[root_pos].post = [Condition(expr=subpost.expr, 
                                               origins=[OriginLoc(isPost=True, index=idx)],
                                               bottom_loc=OriginLoc(index=idx, isPost=True)
                                               ) 
                                        for idx, subpost in enumerate(post)]

        # If there are pre-conditions of constants in the form: A op c, 
        # in which A is a constant symbol, c doesn't include variable symbols, op is in RelExpr.op
        # the pre-conditions will always hold.
        # Add this kind of pre-conditions into assumptions in HCSPs.
        for subpre in self.infos[root_pos].pre:
            if subpre.expr.get_vars().isdisjoint(self.variable_names):
                self.infos[root_pos].assume.append(subpre)

    def set_andR_rule(self, pos, andR):
        if pos not in self.infos:
            self.infos[pos] = CmdInfo()
        self.infos[pos].andR = andR

    def __str__(self):
        res = ""
        for pos, info in self.infos.items():
            res += str(pos) + ":\n"
            res += str(info)
        return res
    
    def print_info(self):
        res = ""
        for pos, info in self.infos.items():
            res += "%s:\n%s" % (pos, info)
        return res

    def simplify_expression(self, e):
        if found_wolfram:
            e = wl_simplify(e, self.functions)

        # Use sympy to simplify
        else:
            e = sp_simplify(e, self.functions)

        return e

    def polynomial_div(self, p, q):
        """Compute the quotient and remainder of polynomial p and q"""
        if found_wolfram:
            quot_remains = wl_polynomial_div(p, q, self.functions)
        else:
            quot_remains = sp_polynomial_div(p, q, self.functions)

        return quot_remains

    def is_polynomial(self, e, constants=set()): 
        """Return True if the given expression is a polynomial, and False otherwise."""
        if found_wolfram:
            result = wl_is_polynomial(e, self.functions, constants)
        else:
            result = sp_is_polynomial(e, self.functions, constants)

        return result

    def compute_variable_set(self, pos=tuple()):
        """Compute variable name set for hcsp program, in which names can be changed by Assign, RandomAssign and ODE"""
        cur_hp = get_pos(self.hp, pos)

        if isinstance(cur_hp, hcsp.Skip):
            pass

        elif isinstance(cur_hp, hcsp.Assign):
            if not isinstance(cur_hp.var_name, expr.AVar):
                raise NotImplementedError
            self.variable_names.add(cur_hp.var_name.name)

        elif isinstance(cur_hp, hcsp.RandomAssign):
            if not isinstance(cur_hp.var_name, expr.AVar):
                raise NotImplementedError
            self.variable_names.add(cur_hp.var_name.name)

        elif isinstance(cur_hp, hcsp.IChoice):
            for i in range(len(cur_hp.hps)):
                sub_pos = pos + (i,)
                self.compute_variable_set(pos=sub_pos)

        elif isinstance(cur_hp, hcsp.Sequence):
            for i in range(len(cur_hp.hps)):
                sub_pos = pos + (i,)
                self.compute_variable_set(pos=sub_pos)

        elif isinstance(cur_hp, hcsp.Loop):
            sub_pos = pos + (0,)
            self.compute_variable_set(pos=sub_pos)

        elif isinstance(cur_hp, hcsp.ODE):
            for name, _ in cur_hp.eqs:
                self.variable_names.add(name)

        elif isinstance(cur_hp, hcsp.ITE):
            for i in range(len(cur_hp.if_hps) + (0 if cur_hp.else_hp is None else 1)):
                sub_pos = pos + (i,)
                self.compute_variable_set(sub_pos)

    def compute_branch_label(self, type, cur_branch=None, sub_blabel=None, post_blabel=None):
        """Compute the branch label for the weakest precondition.
        {P} c {Q}, P is the wp computed by c and Q.

        type: the type of the hcsp program, i.e. c.
        cur_branch: the branch index of the hcsp program, acquired in IChoice, ITE, ODE.
            Example, {P1, P2} c1 ++ c2 {Q}, P1 is computed from c1, P2 is computed from c2,
            then the cur_branch for P1 is 1, the cur_branch for P2 is 2.
        sub_blabel: the branch label of the weakest precondition of sub-hcsp program, acquired in   
            IChoice, ITE, Sequence.
        post_blabel: the branch label of Q. 
                     Note that if P has nothing to do with Q, post_blabel is None by default, for example, the wp, D -> I for ODE.
        """
        if type == 'skip' or type == 'assign' or type == 'randomassign':
            blabel = post_blabel
        
        elif type == 'ichoice' or type == 'ite':
            assert isinstance(cur_branch, int)

            # Case when sub-hp is Skip, Assign or RandomAssign
            if sub_blabel == post_blabel:
                blabel = label.SequenceLabel(label.AtomLabel(cur_branch), post_blabel)

            # Case when sub-hp is IChioce, ITE or ODE
            else:
                assert isinstance(sub_blabel, label.SequenceLabel)
                assert len(sub_blabel.labels) == 2 and sub_blabel.labels[-1] == post_blabel \
                or len(sub_blabel.labels) == 1 # True because all branch labels are built by SequenceLabel(label1, post_blabel), post_blabel could be None
                blabel = label.SequenceLabel(label.NestLabel(cur_branch, sub_blabel.labels[0]), post_blabel)

        elif type == 'ode':
            assert isinstance(cur_branch, str)
            blabel = label.SequenceLabel(label.AtomLabel(cur_branch), post_blabel)

        elif type == 'sequence':
            # The branch label for wp of sequence is equal to 
            # the branch label for wp of the first sub-hp
            blabel = sub_blabel
        
        # Loop is not included because the wp of loop is the invariant, without branch label.
        else:
            raise NotImplementedError

        return blabel     

    def compute_wp(self, *, pos=((),())):
        """Compute weakest-preconditions using the given information."""

        # Obtain the hp at the given position
        cur_hp = get_pos(self.hp, pos[0])
        type = cur_hp.type

        # The post-condition at the given position should already be
        # available.
        # assert pos in self.infos and self.infos[pos].post is not None
        post = self.infos[pos].post

        if isinstance(cur_hp, hcsp.Skip):
            # Skip: {P} skip {P}
            pre = [Condition(
                expr=subpost.expr, 
                path=subpost.path + [pos],
                origins=subpost.origins,
                post_blabel=subpost.blabel,
                blabel=self.compute_branch_label(type=type, post_blabel=subpost.blabel),
                bottom_loc=subpost.bottom_loc, 
                categ=subpost.categ
                ) for subpost in post]
        
        elif isinstance(cur_hp, hcsp.Assign):
            # Assign: {P[e/v]} v := e {P}
            # Compute pre-condition by replacing var_name by expr in the
            # post-condition.
            if not isinstance(cur_hp.var_name, expr.AVar):
                raise NotImplementedError
            if cur_hp.var_name.name in self.constant_names:
                raise NotImplementedError("Constants can not be assigned")

            var = cur_hp.var_name.name
            pre = [Condition(
                expr=subpost.expr.subst({var: cur_hp.expr}), 
                path=subpost.path + [pos],
                origins=subpost.origins,
                post_blabel=subpost.blabel,
                blabel=self.compute_branch_label(type=type, post_blabel=subpost.blabel),
                bottom_loc=subpost.bottom_loc,
                categ=subpost.categ
                ) for subpost in post]
        
        elif isinstance(cur_hp, hcsp.RandomAssign):
            # RandomAssign: replace var_name by var_name_new in post and in cur_hp.expr
            #               pre: cur_hp.expr(newvar/var_name) -> post(newvar/var_name)
            # {(v >= e)[v0/v] -> P[v0/v]} v := {v >= e} {P}
            if not isinstance(cur_hp.var_name, expr.AVar):
                raise NotImplementedError
            if cur_hp.var_name.name in self.constant_names:
                raise NotImplementedError("Constants can not be assigned: {}".format(cur_hp))

            var_str = cur_hp.var_name.name

            # Create a new var
            newvar = create_newvar(var_str, self.names)
            self.names.add(newvar.name)

            #Compute the pre: hp.expr(newvar|var) -> post(newvar|var)
            pre = [Condition(
                expr=expr.imp(cur_hp.expr.subst({var_str: newvar}), subpost.expr.subst({var_str: newvar})), 
                path=subpost.path + [pos],
                origins=subpost.origins,
                post_blabel=subpost.blabel,
                blabel=self.compute_branch_label(type=type, post_blabel=subpost.blabel),
                bottom_loc=subpost.bottom_loc,
                categ=subpost.categ
                ) for subpost in post]

        elif isinstance(cur_hp, hcsp.IChoice):
            # IChoice: 
            #   {P1} c1 {Q}    {P2} c2 {Q}
            # ------------------------------
            #     {P1 && P2} c1 ++ c2 {Q}
            # Pre-condition is the conjunction of the two pre-conditions
            # of subprograms.

            pre = []
            for i in range(len(cur_hp.hps)):
                sub_pos = (pos[0]+(i,), pos[1])
                if sub_pos not in self.infos:
                    self.infos[sub_pos] = CmdInfo()
                sub_info = self.infos[sub_pos]
                sub_info.post = post
                sub_info.assume += self.infos[pos].assume
                sub_info.parent_type = cur_hp.type

                self.compute_wp(pos=sub_pos)

                pre += [Condition(
                    expr=subpre.expr, 
                    path=subpre.path,
                    origins=subpre.origins,
                    post_blabel=subpre.post_blabel,
                    blabel=self.compute_branch_label(type=type, cur_branch=i+1, 
                                                     sub_blabel=subpre.blabel,
                                                     post_blabel=subpre.post_blabel),
                    bottom_loc=subpre.bottom_loc, 
                    categ=subpre.categ
                    ) 
                    for subpre in sub_info.pre]

        elif isinstance(cur_hp, hcsp.ITE):
            # ITE, if b then c1 else c2 endif
            #                       {P1} c1 {Q}  {P2} c2 {Q}  {P3} c3 {Q}
            #-----------------------------------------------------------------------------
            #              {(b1 -> P1) && (!b1 && b2 -> P2)&& (!b1 && !b2 -> P3)} 
            #                      if b1 then c1 elif b2 then c2 else c3 endif 
            #                                         {Q}
            if_hps = cur_hp.if_hps

            sub_imp = []
            if_cond_list = []
            for i in range(len(if_hps) + 1):
                sub_pos = (pos[0] + (i,), pos[1])
                if sub_pos not in self.infos:
                    self.infos[sub_pos] = CmdInfo()
                self.infos[sub_pos].post = post
                self.infos[sub_pos].assume += self.infos[pos].assume
                self.infos[sub_pos].parent_type = cur_hp.type

                if i == len(if_hps) and cur_hp.else_hp is None:
                    sub_pre = post
                else:
                    # Compute the sub-pre-condition for each sub_hp, 
                    # e.g. P1, P2, P3 for c1, c2 and c3
                    self.compute_wp(pos=sub_pos)
                    sub_pre = self.infos[sub_pos].pre

                # Collect the all if conditions.
                if i < len(if_hps):
                    if_cond_list.append(if_hps[i][0])

                # Compute the hypothesis of each implication
                if i == 0:
                    sub_cond = if_cond_list[0]
                elif i < len(if_hps):
                    sub_cond = expr.LogicExpr('&&', expr.neg_expr(expr.list_disj(*if_cond_list[:i])), if_cond_list[i])
                else:
                    sub_cond = expr.neg_expr(expr.list_disj(*if_cond_list[:]))

                # Compute the implicaiton for each if_hp or else_hp
                for subpre in sub_pre:
                    sub_imp.append(
                        Condition(expr=expr.imp(sub_cond, subpre.expr), 
                            path=subpre.path,
                            origins=subpre.origins,
                            post_blabel=subpre.post_blabel,
                            blabel=self.compute_branch_label(type=type, cur_branch=i+1, 
                                                             sub_blabel=subpre.blabel,
                                                             post_blabel=subpre.post_blabel),
                            bottom_loc=subpre.bottom_loc,
                            categ=subpre.categ
                            ))

            pre = sub_imp

        elif isinstance(cur_hp, hcsp.Sequence):
            # Sequence of several commands, apply compute_wp from bottom to top
            cur_post = post
            i_num = 0 # the number of IChoice and ITE in the sub-programs of Sequence.
            for i in reversed(range(len(cur_hp.hps))):
                sub_pos = (pos[0] + (i,), pos[1])
                if sub_pos not in self.infos:
                    self.infos[sub_pos] = CmdInfo()
                sub_info = self.infos[sub_pos]
                sub_info.post = cur_post
                sub_info.assume += self.infos[pos].assume
                sub_info.parent_type = cur_hp.type

                if isinstance(get_pos(self.hp, sub_pos[0]), (hcsp.IChoice, hcsp.ITE)):
                    i_num += 1

                self.compute_wp(pos=sub_pos)
                cur_post = sub_info.pre
            pre = cur_post
            pre = [Condition(expr=subpre.expr, 
                    path=subpre.path,
                    origins=subpre.origins,
                    post_blabel=subpre.post_blabel,
                    blabel=self.compute_branch_label(type=type, sub_blabel=subpre.blabel),
                    bottom_loc=subpre.bottom_loc, 
                    categ=subpre.categ
                    ) for subpre in pre]

        elif isinstance(cur_hp, hcsp.Loop):
            # Loop, currently use the invariant that users offered.

            if cur_hp.constraint != expr.true_expr:
                raise NotImplementedError

            if cur_hp.inv is None:
                raise AssertionError("Loop invariant at position %s is not set." % str(pos))

            for sub_inv in cur_hp.inv:
                if isinstance(sub_inv.expr, expr.LogicExpr):
                    raise NotImplementedError("Logic expression should be split into several relational expressions")

            all_invs = [sub_inv.expr for sub_inv in cur_hp.inv]
            
            # The first time visiting loop program.
            # self.infos[pos].inv is empty by default.
            if not self.infos[pos].inv:

                # Create branches for each invariant.
                # The hp of each branch is still the loop, but with different invariant and assume.
                for i, sub_inv in enumerate(cur_hp.inv):
                    
                    sub_pos = (pos[0], pos[1] + (i,))
                    if sub_pos not in self.infos:
                        self.infos[sub_pos] = CmdInfo()
                    sub_info = self.infos[sub_pos]
                    sub_info.inv = [Condition(expr=sub_inv.expr, 
                                    origins=[OriginLoc(index=i, isInv=True, hp_pos=pos[0])],
                                    bottom_loc=OriginLoc(index=i, isInv=True, hp_pos=pos[0]))]
                    sub_info.assume += self.infos[pos].assume
                                    #   [cur_hp.inv[index].inv for index in range(i)]
                    
                    self.compute_wp(pos=sub_pos)

                # One verification condition is conjunction of sub_inv -> post.
                for i, subpost in enumerate(post):
                    self.infos[pos].vcs.append(
                        Condition(expr=expr.imp(expr.list_conj(*all_invs), subpost.expr), 
                                            path=subpost.path,
                                            blabel=subpost.blabel,
                                            origins=subpost.origins + [OriginLoc(index=i, isInv=True, hp_pos=pos[0]) for i in range(len(cur_hp.inv))],
                                            bottom_loc=subpost.bottom_loc,
                                            categ=subpost.categ,
                                            isVC=True
                                            ))

                pre = [Condition(expr=sub_inv.expr, 
                                 origins=[OriginLoc(index=i, isInv=True, hp_pos=pos[0])],
                                 bottom_loc=OriginLoc(index=i, isInv=True, hp_pos=pos[0]), 
                                 categ="init") \
                       for i, sub_inv in enumerate(cur_hp.inv)]

            # self.infos[pos].inv is set after creating branches.
            # For each branches, we compute wp of loop body and verify that the sub_inv is maintained.
            elif len(self.infos[pos].inv) == 1:
                inv = self.infos[pos].inv[0]

                body_pos = (pos[0] + (0,), pos[1])
                if body_pos not in self.infos:
                    self.infos[body_pos] = CmdInfo()
                body_info = self.infos[body_pos]
                body_info.post = self.infos[pos].inv
                body_info.assume += self.infos[pos].assume
                body_info.parent_type = cur_hp.type
                
                self.compute_wp(pos=body_pos)
                body_pre = body_info.pre

                for sub_pre in body_pre:
                    # bottom_loc add one more tuple to pos to record the annotation index, i.e. invariant index for loop.
                    # Another verification condition is that all invariants together imply the weakest precondition of sub_inv w.r.t the loop body
                    
                    sub_vc = Condition(expr=expr.imp(expr.list_conj(*all_invs), sub_pre.expr),
                                       path=sub_pre.path,
                                       blabel=sub_pre.blabel,
                                       origins=sub_pre.origins + [OriginLoc(index=i, isInv=True, hp_pos=pos[0]) for i in range(len(cur_hp.inv))],
                                       bottom_loc=inv.bottom_loc,
                                       categ="maintain",
                                       isVC=True)

                    self.infos[pos].vcs.append(sub_vc)

                pre = self.infos[pos].inv

        # Reconstruct ODE
        # Only Cut invariant
        elif isinstance(cur_hp, hcsp.ODE):
            # ODE, use differential weakening rule first to prove the hoare triple, 
            # then use other rules to prove invariant.

            # Currently assume out_hp is Skip.
            constraint = cur_hp.constraint
            # The pre-condition computed for invariant and for dw rule is set None initially.
            pre = None
            pre_dw = None


            if cur_hp.out_hp != hcsp.Skip():
                raise NotImplementedError
            for name, _ in cur_hp.eqs:
                if name in self.constant_names:
                    raise NotImplementedError("Constants cannot be changed in ODEs!")

            if not constraint_examination(constraint):
                raise AssertionError("Constriant is supposed to be open intervals!")

            # Construct dictionary corresponding to eqs.
            if not self.infos[pos].eqs_dict:
                for name, deriv in cur_hp.eqs:
                    self.infos[pos].eqs_dict[name] = deriv

            # For the whole ODE program, we use dw rule first, and set the rule for each invariant.
            # For branches generated(when self.infos[pos].dw is False), instead of using dw rule again, 
            # we can use the corresponding rule to verify the invariant directly.
            if self.infos[pos].dw: 
            # TODO: also run if no invariants are specified? testVerify62 testVerify54 testVerify53 testVerify52 testVerify50 testVerify55

                if cur_hp.inv is None:
                    cur_hp.inv = (assertion.CutInvariant(expr=expr.true_expr, proof_methods=None),)

                assert all(isinstance(inv, assertion.CutInvariant) for inv in cur_hp.inv)

                # Apply differential weakening (dW)

                # dW Rule (always applied automatically)
                #   [[D]] <x_dot = f(x)> [[I]]      I && Boundary of D -> Q  
                #-----------------------------------------------------------------------
                #           {(D -> I) && (!D -> Q)} <x_dot = f(x) & D> {Q}
                
                # dWT: Case when no invariant is offered(I is set as true)
                #                 Boundary of D -> Q  
                #-----------------------------------------------------------------------
                #           {!D -> Q} <x_dot = f(x) & D> {Q}
                pre_dw = [Condition(expr=expr.imp(constraint, inv.expr), 
                                    origins=[OriginLoc(index=i, isInv=True, hp_pos=pos[0]), 
                                             OriginLoc(isConstraint=True, hp_pos=pos[0])],
                                    categ="init",
                                    bottom_loc=OriginLoc(index=i, isInv=True, hp_pos=pos[0]),
                        ) for i, inv in enumerate(cur_hp.inv) if inv.expr is not expr.true_expr] \
                        + \
                         [Condition(expr=expr.imp(expr.neg_expr(constraint), 
                                                     subpost.expr),
                                    path=subpost.path,
                                    post_blabel=subpost.blabel,
                                    blabel=self.compute_branch_label(type=type, cur_branch='skip', post_blabel=subpost.blabel),
                                    origins=subpost.origins + [OriginLoc(isConstraint=True, hp_pos=pos[0])],
                                    bottom_loc=subpost.bottom_loc,
                                    categ=subpost.categ
                                    )
                          for subpost in post]
                boundary = compute_boundary(constraint)

                inv_exprs = [inv.expr for inv in cur_hp.inv]
                inv_conj = expr.conj(*inv_exprs)

                # When I is false_expr, (I && Boundary of D -> Q) is true_expr, which can be omitted. 
                if inv_conj is not expr.false_expr:
                    for subpost in post:
                        e = expr.imp(expr.conj(inv_conj, boundary), subpost.expr)
                        self.infos[pos].vcs.append(
                            Condition(
                                expr=e, 
                                path=subpost.path,
                                blabel=subpost.blabel,
                                categ=subpost.categ,
                                origins=subpost.origins + 
                                [OriginLoc(index=i, isInv=True, hp_pos=pos[0]) for i in range(len(cur_hp.inv))] + [OriginLoc(isConstraint=True, hp_pos=pos[0])],
                                bottom_loc=subpost.bottom_loc,
                                isVC=True
                                ))
                invs = []
                for i, inv in enumerate(cur_hp.inv):
                    sub_pos = (pos[0], pos[1] + (i,))
                    if sub_pos not in self.infos:
                        self.infos[sub_pos] = CmdInfo()

                    self.infos[sub_pos].assume = self.infos[pos].assume + invs[:i]
                    self.infos[sub_pos].eqs_dict = self.infos[pos].eqs_dict
                    self.infos[sub_pos].inv = Condition(expr=inv.expr, 
                                              origins=[OriginLoc(index=i, isInv=True, hp_pos=pos[0])], 
                                              bottom_loc=OriginLoc(index=i, isInv=True, hp_pos=pos[0]))
                    invs.append(self.infos[sub_pos].inv)
                    self.infos[sub_pos].post = [Condition(expr=inv.expr, 
                                                origins=[OriginLoc(index=i, isInv=True, hp_pos=pos[0])], 
                                                bottom_loc=OriginLoc(index=i, isInv=True, hp_pos=pos[0]))]

                    if inv.rule is None and inv.expr in (expr.true_expr, expr.false_expr):
                            self.infos[sub_pos].tv = True #TODO: Use the name "tv"(trival)?
                            self.infos[sub_pos].dw = False
                            assert inv.rule_arg is None
                    elif inv.rule == "di" or \
                    (inv.rule is None and inv.expr not in (expr.true_expr, expr.false_expr)):
                        self.infos[sub_pos].dI_rule = True
                        self.infos[sub_pos].dw = False
                        assert inv.rule_arg is None
                    elif inv.rule == "dbx":
                        self.infos[sub_pos].dbx_rule = True
                        self.infos[sub_pos].dw = False
                        if inv.rule_arg is not None:
                            self.infos[sub_pos].dbx_cofactor = inv.rule_arg
                    elif inv.rule == "bc":
                        self.infos[sub_pos].barrier_rule = True
                        self.infos[sub_pos].dw = False
                        assert inv.rule_arg is None              
                    elif inv.rule == "sln":
                        self.infos[sub_pos].sln_rule = True
                        self.infos[sub_pos].dw = False
                        assert inv.rule_arg is None
                    else:
                        if inv.rule is not None:
                            raise NotImplementedError("Unknown ODE method")

                    self.compute_wp(pos=sub_pos)


            # post_conj = expr.conj(*[vc.expr for vc in post])

            # Only use the boundary.
            elif self.infos[pos].tv:

                assert self.infos[pos].inv.expr in (expr.false_expr, expr.true_expr), \
                    "Invariant should be true or false!"

            # Use solution axiom
            # 
            #             P ->
            # ForAll t >= 0  ((ForAll 0 <= s < t D(y(s)) && not D(y(t))) -> (ForAll 0 <= s <= t Q(y(s)))
            #--------------------------------------------------------------------------------------------
            #      {P} <x_dot = f(x) & D(x)> {Q(x)}
            #
            # Assume y(.) solve the symbolic initial value problem y'(t) = f(y), y(0) = x

            elif self.infos[pos].sln_rule:

                # Create a new variable to represent time
                time_var = create_newvar('t', self.names)
                self.names.add(time_var.name)

                # Solution is, e.g. {'x' : x + v*t + a*t^2/2, 'v' : v + a*t}.
                solution_dict = solveODE(cur_hp, self.names, time_var.name)

                # Create a new variable to represent 's' in 'ForAll 0 <= s < t'
                in_var = create_newvar('s', self.names)
                self.names.add(in_var.name)

                # Compute y_s, alias y(s)
                # e.g. {'x' : x + v*s + a*s^2/2, 'v' : v + a*s
                y_s = dict()
                for fun_name, sol in solution_dict.items():
                    sol_s = sol.subst({time_var.name : in_var})
                    y_s[fun_name] = sol_s

                D_y_t = constraint.subst(solution_dict)
                D_y_s = constraint.subst(y_s)
                Q_y_s = self.infos[pos].inv.expr.subst(y_s)

                # Compute the hypothesis of implication
                # ForAll (s, 0 <= s < t -> D(y(s)) && not D(y(t))
                sub_cond = expr.ForAllExpr(in_var.name, 
                                expr.imp(expr.LogicExpr('&&', 
                                                        expr.RelExpr('<=', expr.AConst(0), in_var),
                                                        expr.RelExpr('<', in_var, time_var)),
                                         D_y_s))
                cond = expr.LogicExpr('&&', 
                                      sub_cond,
                                      expr.LogicExpr('!', D_y_t))
                # Compute the conclusion of implication
                # ForAll (s, 0 <= s <= t -> Q(y(s))
                conclu = expr.ForAllExpr(in_var.name,
                                expr.imp(expr.LogicExpr('&&', 
                                                        expr.RelExpr('<=', expr.AConst(0), in_var),
                                                        expr.RelExpr('<=', in_var, time_var)),
                                         Q_y_s))

                pre = [Condition(expr=expr.ForAllExpr(time_var.name,
                                                    expr.imp(expr.RelExpr('>=', time_var, expr.AConst(0)),
                                                             expr.imp(cond, conclu))),
                                path=[pos],
                                origins=self.infos[pos].inv.origins,
                                categ="maintain",
                                bottom_loc=self.infos[pos].inv.bottom_loc)]                                         

            # Use dI rules
            # When dI_inv is set or by default
            elif self.infos[pos].dI_rule: 

                dI_inv = self.infos[pos].inv.expr          
                # Compute the differential of inv.
                # Compute the boundary of constraint. 
                # One semi-verification condition is boundary of constraint -> differential of inv.
                differential = compute_diff(dI_inv, eqs_dict=self.infos[pos].eqs_dict, functions=self.functions)
                vc = expr.imp(constraint, differential)
     
                self.infos[pos].vcs.append(Condition(expr=vc, 
                                                     path=[pos],
                                                     origins=self.infos[pos].inv.origins,
                                                     categ="maintain", 
                                                     bottom_loc=self.infos[pos].inv.bottom_loc,
                                                     isVC=True))


            # Using dbx rule
            # Cases when dbx_inv is "e == 0".
                # Use Darboux Equality Rule
                #          D -> e_lie_deriv == g * e
                #--------------------------------------------    (g is the cofactor)
                #   {e == 0} <x_dot = f(x) & D> {e == 0}
            # Cases when dbx_inv is e >(>=) 0.
                # Use Darboux Inequality Rule.
                #           D -> e_lie_deriv >= g * e
                # ---------------------------------------------
                #    e >(>=) 0 <x_dot = f(x) & D> e >(>=) 0
            elif self.infos[pos].dbx_rule:

                dbx_inv = self.infos[pos].inv.expr
                bottom_loc = self.infos[pos].inv.bottom_loc
                origins = self.infos[pos].inv.origins

                # For example, simplify !(x > 1) to x <= 1
                if isinstance(dbx_inv, expr.LogicExpr): 
                    dbx_inv = self.simplify_expression(dbx_inv)
                
                assert isinstance(dbx_inv, expr.RelExpr) and \
                       dbx_inv.op in {'==', '>', '>=', '<', '<='}, \
                    print("The wrong type of %s is %s" %(dbx_inv, type(dbx_inv)))

                # Tranlate '<' and '<=' into '>' and '>='.
                if dbx_inv.op == '<':
                    dbx_inv = expr.RelExpr('>', dbx_inv.expr2, dbx_inv.expr1)
                elif dbx_inv.op == '<=':
                    dbx_inv = expr.RelExpr('>=', dbx_inv.expr2, dbx_inv.expr1)
                # Translate into the form e == 0 or e >(>=) 0
                if dbx_inv.expr2 != expr.AConst(0):
                    expr1 =  expr.OpExpr('-', dbx_inv.expr1, dbx_inv.expr2)
                    expr1 = self.simplify_expression(expr1)
                    dbx_inv = expr.RelExpr(dbx_inv.op, expr1, expr.AConst(0)) 
                
                e = dbx_inv.expr1
                # Compute the lie derivative of e.
                e_lie_deriv = compute_diff(e, eqs_dict=self.infos[pos].eqs_dict, functions=self.functions)             

                # Cases when dbx_inv is "e == 0".
                # Use Darboux Equality Rule
                #          D -> e_lie_deriv == g * e
                #--------------------------------------------    (g is the cofactor)
                #   {e == 0} <x_dot = f(x) & D> {e == 0}
                if dbx_inv.op == "==" :

                    # Case when the cofactor g is not offered.
                    # Compute the cofactor g by auto.
                    if self.infos[pos].dbx_cofactor is None:
                        g = expr.OpExpr('/', e_lie_deriv, e)
                        g = self.simplify_expression(g)

                        assert self.is_polynomial(g, self.constant_names) is True
                        self.infos[pos].dbx_cofactor = g

                    # Case when cofactor g is offered by the user.
                    else:
                        g = self.infos[pos].dbx_cofactor
                        assert self.is_polynomial(g, self.constant_names) is True

                        # Boundary of D -> e_lie_deriv == g * e
                        vc = expr.imp(constraint, expr.RelExpr('==', e_lie_deriv, 
                                                                    expr.OpExpr('*', g, e)))

                        self.infos[pos].vcs.append(Condition(expr=vc, 
                                                             path=[pos],
                                                             origins=origins, 
                                                             bottom_loc=bottom_loc,
                                                             categ="maintain", 
                                                             isVC=True))

                # Cases when dbx_inv is e >(>=) 0.
                # Use Darboux Inequality Rule.
                #           D -> e_lie_deriv >= g * e
                # ---------------------------------------------
                #    e >(>=) 0 <x_dot = f(x) & D> e >(>=) 0
                elif dbx_inv.op in {'>', '>='}:
                    # Cases when the cofactor g is not offered.
                    # Compute the cofactor automatically.
                    if self.infos[pos].dbx_cofactor is None:
                        quot_remains = self.polynomial_div(e_lie_deriv, e)
                        
                        # Several quot and remain pairs may returned from the polynomial division.
                        # If there is a remain >= 0, there exists a quot satisfying g in dbx_rule.
                        vc_comps = []
                        for _, remain in quot_remains.items():
                            # remain (e_lie_deriv - g * e) >= 0.
                            vc_comp = expr.RelExpr('>=', remain, expr.AConst(0))
                            vc_comps.append(vc_comp)
                        vc = expr.imp(constraint, expr.list_disj(*vc_comps))

                        self.infos[pos].vcs.append(Condition(expr=vc, 
                                                            path=[pos], 
                                                            origins=origins, 
                                                            categ="maintain",
                                                            bottom_loc=bottom_loc, 
                                                            isVC=True))

                    # Cases when the cofactor g is offered by the user.
                    else:
                        g = self.infos[pos].dbx_cofactor
                        assert self.is_polynomial(g, self.constant_names) is True

                        vc = expr.imp(constraint, expr.RelExpr('>=', e_lie_deriv, 
                                                                    expr.OpExpr('*', self.infos[pos].dbx_cofactor, e)))
                        self.infos[pos].vcs.append(Condition(expr=vc, path=[pos], origins=origins, categ="maintain", bottom_loc=bottom_loc, isVC=True))              
          

            # Use barrier certificate
            #             D && e == 0 -> e_lie > 0
            # --------------------------------------------------
            #      {e >=(>) 0} <x_dot = f(x) & D> {e >=(>) 0}
            elif self.infos[pos].barrier_rule:

                barrier_inv = self.infos[pos].inv.expr
                bottom_loc = self.infos[pos].inv.bottom_loc
                origins = self.infos[pos].inv.origins

                if isinstance(barrier_inv, expr.LogicExpr):
                    barrier_inv = self.simplify_expression(barrier_inv)

                assert isinstance(barrier_inv, expr.RelExpr) and \
                       barrier_inv.op in {'<=', '>=', '>', '<'}

                # TODO: e should be continous

                # Translate '<=' into '>='
                if barrier_inv.op == '<=':
                    barrier_inv = expr.RelExpr('>=', barrier_inv.expr2, barrier_inv.expr1)
                elif barrier_inv.op == '<':
                    barrier_inv = expr.RelExpr('>', barrier_inv.expr2, barrier_inv.expr1)

                # Translate 'e >= c' into 'e - c >= 0'
                if barrier_inv.expr2 != 0:
                    expr1 = expr.OpExpr('-', barrier_inv.expr1, barrier_inv.expr2)
                    # expr1 = self.simplify_expression(expr1)
                    barrier_inv = expr.RelExpr(barrier_inv.op, expr1, expr.AConst(0))

                e = barrier_inv.expr1
                e_lie = compute_diff(e, eqs_dict=self.infos[pos].eqs_dict, functions=self.functions)

                # vc: D && e == 0 -> e_lie > 0
                vc = expr.imp(expr.LogicExpr('&&', constraint, 
                                                   expr.RelExpr('==', e, expr.AConst(0))),
                              expr.RelExpr('>', e_lie, expr.AConst(0)))

                self.infos[pos].vcs.append(Condition(expr=vc, path=[pos], origins=origins, categ="maintain", bottom_loc=bottom_loc, isVC=True))


            else:
                raise AssertionError("No invariant set at position %s." % str(pos))

            if pre is not None and pre_dw is not None:
                pre = pre + pre_dw
            elif pre_dw is not None:
                pre = pre_dw
            # If pre is None and pre_dw is None, no pre is computed, so pre is still None.

        else:
            raise NotImplementedError

        # Assign pre-condition, or create a new verification condition if the
        # pre-condition is already set.
        if self.infos[pos].pre is None:
            self.infos[pos].pre = pre
        else:
            # self.infos[pos].pre is the pre-condition of the whole hp.
            pre_conj = expr.conj(*[subpre.expr for subpre in self.infos[pos].pre])
            subpre_origins = []
            for subpre in self.infos[pos].pre:
                subpre_origins.extend(subpre.origins)
            for i, vc in enumerate(pre):
                self.infos[pos].vcs.append(
                    Condition(expr=expr.imp(pre_conj, vc.expr), 
                              path=vc.path,
                              blabel=vc.blabel, 
                              origins=vc.origins + subpre_origins,
                              bottom_loc=vc.bottom_loc, 
                              categ=vc.categ,
                              isVC=True
                              ))

        # Add assume into the hypothesis of every verification condition.
        if self.infos[pos].assume:
            assume = expr.list_conj(*[sub_assume.expr for sub_assume in self.infos[pos].assume])
            asm_origins = []
            for sub_assume in self.infos[pos].assume:
                asm_origins.extend(sub_assume.origins)
            
            for i in range(len(self.infos[pos].vcs)):
                vc = self.infos[pos].vcs[i]
                self.infos[pos].vcs[i] = Condition(
                    expr=expr.imp(assume, vc.expr), 
                    path=vc.path,
                    blabel=vc.blabel,
                    origins=vc.origins + asm_origins,
                    bottom_loc=vc.bottom_loc, 
                    categ=vc.categ,
                    isVC=True
                    )
        
    def convert_imp(self, e):
        """Convert implication from (p -> q -> u) to (p && q) -> u,
        in which the right expression won't be an implication """
        if isinstance(e, expr.LogicExpr) and e.op == '->':
            if isinstance(e.exprs[1], expr.LogicExpr) and e.exprs[1].op == '->':
                l_expr = expr.LogicExpr('&&', e.exprs[0], e.exprs[1].exprs[0])
                r_expr = e.exprs[1].exprs[1]
                return self.convert_imp(expr.imp(l_expr, r_expr))
            # p -> q
            else:
                return e
        else:
            return e
 
    def set_solver(self, vc):
        """Set the solver of verification condition, which is z3 by default, as the one indicated in  proof method"""
        assert isinstance(vc, Condition) and vc.isVC is True
        assert vc.bottom_loc is not None

        isPost = vc.bottom_loc.isPost

        # Proof method of each vc is stored in the bottom-most assertion related.

        if isPost:
            assertion = self.post[vc.bottom_loc.index]
        
        else:
            assert vc.bottom_loc.hp_pos is not None
            bottom_hp = get_pos(self.hp, vc.bottom_loc.hp_pos)

            assert isinstance(bottom_hp, (hcsp.Loop, hcsp.ODE))
            assertion = bottom_hp.inv[vc.bottom_loc.index]

        # Match the computed label of vc and the one in proof method.
        for pms in assertion.proof_methods.pms:
            if str(vc.comp_label) == str(pms.label):
                vc.solver = pms.method

        return vc

    def get_all_vcs(self):
        all_vcs = dict()
        for pos, info in self.infos.items():   
            if info.vcs:
                all_vcs[pos] = info.vcs
        return all_vcs

    def verify(self):
        """Verify all VCs in self."""
        all_vcs = self.get_all_vcs()
        
        for pos, vcs in all_vcs.items():
            for vc in vcs:
                vc = self.set_solver(vc)
                if not self.verify_vc(vc):
                    print("The failed verification condition is :\n", pos, ':', str(vc))
                    return False
        return True



    def verify_vc(self, vc):
        """Verify one verfication condition with its solver"""
        solver = vc.solver
        if solver == 'wolfram':
            if wl_prove(vc.expr, self.functions):
                return True
            else:
                return False

        elif solver == 'z3':
            if z3_prove(vc.expr, self.functions):
                return True
            else:
                return False

        else:
            raise AssertionError("Please choose an arithmetic solver.")




            





from ss2hcsp.hcsp.parser import expr_parser, hp_parser
from ss2hcsp.hcsp import hcsp as hp
from ss2hcsp.hcsp.expr import RelExpr,LogicExpr,FunExpr
import re


class SF_State:
    """Represents a state in a Stateflow diagram.

    Each state has an ID and a name.

    Each state records events (represented by its name) that
    happens when:

    - entry: when first entering the state.
    - during: when none of the outgoing edges can be executed.
    - exit: when exiting the state.

    """
    def __init__(self, ssid, inner_trans=(), name="", en=None, du=None, ex=None):
        self.ssid = ssid  # ID of the state
        self.name = name  # Name of the state
        self.en = en  # entry event
        self.du = du  # during event
        self.ex = ex  # exit event
        self.father = None  # parent state
        self.children = list()  # list of children states
        self.chart = None  # The chart containing this state
        self.whole_name=""

        # Inner transitions of this state
        assert isinstance(inner_trans, (list, tuple))
        self.inner_trans = inner_trans

    def get_state_whole_name(self):
        if isinstance(self, (OR_State, AND_State)):
            s = self.name
            if self.father is None:
                return self.name
            else:
                child_s = self.father.get_state_whole_name()
                s = child_s + "_" + s
            return s

    def __eq__(self, other):
        return self.ssid == other.ssid

    def __str__(self):
        # Header: ID and name
        if isinstance(self, OR_State):
            result = "OR"
        elif isinstance(self, AND_State):
            result = "AND"
        result += "{id=" + self.ssid + ", name=" + self.name + "}\n"

        # Display the default transition
        if isinstance(self, OR_State) and self.default_tran:
            result += "  Default: " + str(self.default_tran) + "\n"

        # Display the history junction
        if isinstance(self, OR_State) and self.has_history_junc:
            result += "  History: " + str(self.has_history_junc) + "\n"

        # Events
        if self.en:
            result += "  en: " + str(self.en) + "\n"
        if self.du:
            result += "  du: " + str(self.du) + "\n"
        if self.ex:
            result += "  ex: " + str(self.ex) + "\n"

        # Display output transitions
        if isinstance(self, OR_State) and self.out_trans:
            result += "  Out-Transitions:\n"
            for tran in self.out_trans:
                result += "    " + str(tran) + "\n"

        # Display inner transitions
        if self.inner_trans:
            result += "  IN-transitions:\n"
            for tran in self.inner_trans:
                result += "    " + str(tran) + "\n"

        # Display children
        if self.children:
            result += "  Children:\n"
            for child in self.children:
                child_lines = str(child).splitlines()
                result += "\n".join("    " + line for line in child_lines) + "\n"

        return result

    def get_all_descendants(self):
        """Returns a dictionary containing all descendents of self.
        
        Returns - dict(str, SF_State): mapping from ID to nodes.

        """
        assert isinstance(self, (AND_State, OR_State))
        descendants = dict()
        for child in self.children:
            assert child.ssid not in descendants
            descendants[child.ssid] = child
            if isinstance(child, (AND_State, OR_State)):
                child_descendants = child.get_all_descendants()
                #assert all(ssid not in descendants for ssid in child_descendants.keys())
                descendants.update(child_descendants)
        return descendants

    def get_vars(self):
        """Returns set of variables in the actions of the current state.
        
        This function enters recursively into child states.

        """
        var_set = set()

        # Collect variables in entry, during and exit actions
        en_du_ex_acts = (self.en if self.en else list()) \
                        + (self.du if self.du else list()) \
                        + (self.ex if self.ex else list())
        for act in en_du_ex_acts:
            #assert isinstance(act, hp.HCSP)
            if isinstance(act, (hp.Assign, hp.Sequence)):
                var_set = var_set.union(act.get_vars())

        # All states have inner transitions.
        for tran in self.inner_trans:
            var_set = var_set.union(tran.get_vars())

        # Only OR states have outgoing and default transitions.
        if isinstance(self, OR_State):                           
            for tran in self.out_trans:
                var_set = var_set.union(tran.get_vars())
            if self.default_tran:
                var_set = var_set.union(self.default_tran.get_vars())
        
        # Variables in child states.
        for child in self.children:
            var_set = var_set.union(child.get_vars())

        return var_set

    def check_children(self):
        """Check validity of state.

        Checks:
        - if a state contains AND states, then it should not contain OR states
          or junctions.

        """
        has_AND_state = has_OR_state = has_Junction = False
        for child in self.children:
            if isinstance(child, AND_State):
                has_AND_state = True
            elif isinstance(child, OR_State):
                has_OR_state = True
            elif isinstance(child, Junction):
                has_Junction = True
            else:  # Error State
                return False

        # AND_state cannot be in the same father state with OR_state or Junction
        if has_AND_state and (has_OR_state or has_Junction):
            return False

        # Check child states.
        for child in self.children:
            if isinstance(child, (AND_State, OR_State)) and not child.check_children():
                return False

        return True


class OR_State(SF_State):
    """Represents an OR state."""
    def __init__(self, ssid, out_trans=(), inner_trans=(), name="",
                 en=None, du=None, ex=None, default_tran=None):
        super(OR_State, self).__init__(ssid, inner_trans, name, en, du, ex)
        self.out_trans = out_trans        # outgoing transitions
        self.default_tran = default_tran  # default transition to this state
        self.has_history_junc = False     # history junction to this state


class AND_State(SF_State):
    """Represents an AND state."""
    def __init__(self, ssid, inner_trans=(), name="", en=None, du=None, ex=None, order=0):
        super(AND_State, self).__init__(ssid, inner_trans, name, en, du, ex)
        self.order = order                # order within the parent state 


class Junction:
    """Represents a junction."""
    def __init__(self, ssid, out_trans, name="", junc_type="", default_tran=None):
        self.ssid = ssid
        self.out_trans = out_trans
        self.name = name
        self.whole_name = ""
        self.father = None
        self.visited = False
        self.type = junc_type
        self.processes = list()
        self.tran_acts = list()  # the queue to store transition actions
        self.default_tran = default_tran
        self.is_parse_act = False

    def __str__(self):
        result = "JUN(" + self.ssid + ") " + self.name + "\n"
        for tran in self.out_trans:
            result += "  " + str(tran) + "\n"
        return result

    def get_vars(self):
        var_set = set()
        for tran in self.out_trans:
            var_set = var_set.union(tran.get_vars())
        return var_set


class GraphicalFunction:
    def __init__(self, name, params, return_var, transitions, junctions):
        self.name = name
        self.params = params
        self.return_var = return_var
        self.transitions = transitions
        self.junctions = junctions

        # Obtain default transition
        self.default_tran = None
        for ssid, tran in self.transitions.items():
            if tran.src is None:
                self.default_tran = tran
                break
        assert self.default_tran, "GraphicalFunction: no default transition found"

    def __str__(self):
        res = "GraphicalFunction(%s,%s,%s\n" % (self.name, self.params, self.return_var)
        res += "  Junctions:\n"
        for junc in self.junctions:
            for line in str(junc).split('\n'):
                res += "  " + line + "\n"
        res += "  Default transition:\n"
        res += "    " + str(self.default_tran) + "\n"
        res += ")"
        return res



"""Simulink diagrams."""

import lark
from decimal import Decimal

from ss2hcsp.sl.sl_line import SL_Line
from ss2hcsp.sl.sl_block import get_gcd
from ss2hcsp.matlab import function, convert
from ss2hcsp.matlab import function
from ss2hcsp.matlab.parser import expr_parser, function_parser, \
    transition_parser, func_sig_parser, state_op_parser
from ss2hcsp.matlab import convert
from ss2hcsp.sl.Continuous.clock import Clock
from ss2hcsp.sl.port import Port
from ss2hcsp.sl.Continuous.integrator import Integrator
from ss2hcsp.sl.Continuous.constant import Constant
from ss2hcsp.sl.Continuous.signalBuilder import SignalBuilder
from ss2hcsp.sl.Continuous.source import Sine
from ss2hcsp.sl.MathOperations.product import Product
from ss2hcsp.sl.MathOperations.bias import Bias
from ss2hcsp.sl.MathOperations.gain import Gain
from ss2hcsp.sl.MathOperations.add import Add
from ss2hcsp.sl.MathOperations.my_abs import Abs
from ss2hcsp.sl.MathOperations.sqrt import Sqrt
from ss2hcsp.sl.MathOperations.square import Square
from ss2hcsp.sl.LogicOperations.logic import And, Or, Not
from ss2hcsp.sl.LogicOperations.relation import Relation
from ss2hcsp.sl.LogicOperations.reference import Reference
from ss2hcsp.sl.SignalRouting.switch import Switch
from ss2hcsp.sl.SignalRouting.scope import Scope
from ss2hcsp.sl.SubSystems.subsystem import Subsystem, Triggered_Subsystem, Enabled_Subsystem
from ss2hcsp.sl.Discontinuities.saturation import Saturation
from ss2hcsp.sl.Discrete.unit_delay import UnitDelay 
from ss2hcsp.sl.Discrete.DiscretePulseGenerator import DiscretePulseGenerator
from ss2hcsp.sl.Discrete.discrete_PID_controller import DiscretePID
from ss2hcsp.sl.MathOperations.min_max import MinMax
from ss2hcsp.sf.sf_state import AND_State, OR_State, Junction, GraphicalFunction
from ss2hcsp.sf.sf_chart import SF_Chart
from ss2hcsp.sf.sf_transition import Transition
from ss2hcsp.sf.sf_message import SF_Message,SF_Data
from ss2hcsp.sf.sf_event import SF_Event
from ss2hcsp.sl.discrete_buffer import Discrete_Buffer
from ss2hcsp.sl.mux.mux import Mux
from ss2hcsp.sl.DataStore.DataStore import DataStoreMemory, DataStoreRead
from xml.dom.minidom import parse, Element
from xml.dom.minicompat import NodeList
import operator


def parse_sample_time(sample_time):
    """Convert sample time in string to integer or decimal."""
    if sample_time is None or sample_time == 'inf':
        return -1
    elif '.' in sample_time:
        v = Decimal(sample_time)
        return int(v * 1000) / Decimal(1000)
    else:
        return int(sample_time)

def parse_value(value, default=None):
    """Parse value to integer or float."""
    if value:
        if '.' in value:
            return float(value)
        else:
            return int(value)
    else:
        return default

def replace_spaces(name):
    """Replace spaces and newlines in name with underscores."""
    return name.strip().replace(' ', '_').replace('\n', '_')

def get_attribute_value(block, attribute, name=None):
    for node in block.getElementsByTagName("P"):
        if node.getAttribute("Name") == attribute:
            if node.childNodes:
                if not name or name == node.childNodes[0].data:
                    return node.childNodes[0].data
    return None


class SL_Diagram:
    """Represents a Simulink diagram."""
    def __init__(self, location=""):
        # Dictionary of blocks indexed by name
        self.blocks_dict = dict()

        # Dictionary of STATEFLOW parameters indexed by name
        self.chart_parameters = dict()

        # XML data structure
        self.model = None

        # Name of the diagram, set by parse_xml
        self.name = None

        # Variables that needs to display
        self.outputs = list()

        # Different parts of the diagram
        self.continuous_blocks = list()
        self.discrete_blocks = list()
        self.scopes = list()
        self.dsms = list()

        self.others = list()
        
        # Parsed model of the XML file
        if location:
            with open(location) as f:
                self.model = parse(f)

    def parse_stateflow_xml(self):
        """Parse stateflow charts from XML."""

        def get_acts(acts):
            """Convert action into a sequence of actions."""
            lists = list()
            if isinstance(acts, function.Sequence):
                for act in [acts.cmd1, acts.cmd2]:
                    if isinstance(act, function.Sequence):
                        lists.extend(get_acts(act))
                    else:
                        lists.append(act)
            else:
                lists.append(acts)
            return lists

        def get_transitions(blocks):
            """Obtain the list of transitions.
            
            Returns a dictionary mapping transitions IDs to transition objects.
            
            """
            assert isinstance(blocks, NodeList)

            tran_dict = dict()
            for block in blocks:
                if block.nodeName == "transition":
                    # Obtain transition ID, label, execution order
                    tran_ssid = block.getAttribute("SSID")
                    tran_label = get_attribute_value(block, "labelString")
                    if tran_label is None:
                        tran_label = function.TransitionLabel(None, None, None, None)
                    else:
                        try:
                            tran_label = transition_parser.parse(tran_label)
                        except lark.exceptions.UnexpectedToken as e:
                            print("When parsing transition label", tran_label)
                            raise e
                    order = int(get_attribute_value(block, "executionOrder"))

                    # Each transition must have exactly one source and destination
                    assert len([child for child in block.childNodes if child.nodeName == "src"]) == 1
                    assert len([child for child in block.childNodes if child.nodeName == "dst"]) == 1
                    src_ssid, dst_ssid = None, None
                    for child in block.childNodes:
                        if child.nodeName == "src":
                            src_ssid = get_attribute_value(child, "SSID")
                        elif child.nodeName == "dst":
                            dst_ssid = get_attribute_value(child, "SSID")

                    # There should be no repeated transition IDs
                    assert tran_ssid not in tran_dict
                    tran_dict[tran_ssid] = Transition(ssid=tran_ssid, label=tran_label, order=order,
                                                      src=src_ssid, dst=dst_ssid)
            return tran_dict

        all_out_trans = dict()

        def get_children(block):
            """Get lists of children states and junctions of the current
            block.

            Returned value is:

            _states: list of SF_State objects.
            _junction: list of Junction objects.
            _functions : list of Function objects.
            
            """
            assert isinstance(block, Element)

            children = [child for child in block.childNodes if child.nodeName == "Children"]
            if not children:
                return list(), list(), list()
            assert len(children) == 1

            _states, _junctions, _functions = list(), list(), list()

            # Get list of outgoing transitions from states in children
            out_trans_dict = get_transitions(children[0].childNodes)
            
            # The number of default transitions is at most one
            assert len([tran for tran in out_trans_dict.values() if tran.src is None]) <= 1

            # Now traverse each child node. We consider two cases of child nodes:
            # states and junctions.
            for child in children[0].childNodes:
                if child.nodeName == "state":
                    ssid = child.getAttribute("SSID")
                    state_type = get_attribute_value(child, "type")

                    # There are three cases of states: FUNC_STATE (definition of functions),
                    # AND_STATE and OR_STATE.
                    if state_type == "FUNC_STATE":
                        # Extract functions
                        fun_name = get_attribute_value(child, "labelString")
                        fun_script = get_attribute_value(child, "script")
                        if fun_script:
                            # Has script, directly use parser for matlab functions
                            fun_name=function_parser.parse(fun_script).name
                            hp= convert.convert_cmd(function_parser.parse(fun_script).cmd)
                            ru=function_parser.parse(fun_script).return_var
                            exprs=function.ListExpr(function_parser.parse(fun_script).params) if function_parser.parse(fun_script).params is not None else None
                            fun_type="MATLAB_FUNCTION"
                            if isinstance(ru,(function.Var,function.FunctionCall)):
                                return_var=ru
                            elif isinstance(ru,tuple):
                                return_var=function.ListExpr(*ru)
                            _functions.append(function_parser.parse(fun_script))
                            # _functions.append(Function(fun_name,exprs,ru,hp,None,fun_type))
                        else:
                            fun_type="GRAPHICAL_FUNCTION"
                            children = [c for c in child.childNodes if c.nodeName == "Children"]
                            assert len(children) == 1
                            sub_trans = get_transitions(children[0].childNodes)
                            sub_states, sub_junctions, sub_functions = get_children(child)
                            assert len(sub_states) == 0 and len(sub_functions) == 0

                            try:
                                fun_name, fun_params, fun_return = func_sig_parser.parse(fun_name)
                            except lark.exceptions.UnexpectedToken as e:
                                print("When parsing function signature", fun_name)
                                raise e

                            chart_state1 = OR_State(ssid=ssid, name=fun_name)
                            for state in  sub_junctions:
                                state.father = chart_state1
                                chart_state1.children.append(state)
                            # _states.append(chart_state1)
                            hp=None
                            graph_fun = GraphicalFunction(fun_name, fun_params, fun_return, sub_trans, sub_junctions)
                            _functions.append(graph_fun)
                            # _functions.append(Function(fun_name,fun_params,fun_return,hp,chart_state1,fun_type))

                    elif state_type in ("AND_STATE", "OR_STATE"):
                        # Extract AND- and OR-states

                        # The format of state labels is as follows:
                        # First line is the name of the state, the remaining lines
                        # specify en, du, and ex actions.
                        labels = get_attribute_value(child, "labelString")
                        label = state_op_parser.parse(labels)
                        name = str(label.name)
                        
                        # Get en, du and ex actions
                        en = get_acts(label.en_op.op) if label.en_op is not None else []
                        du = get_acts(label.du_op.op) if label.du_op is not None else []
                        ex = get_acts(label.ex_op.op) if label.ex_op is not None else []
                        
                        # Get default_tran and out_trans
                        default_tran = None
                        out_trans = list()
                        dictMerged2 = dict(out_trans_dict)
                        dictMerged2.update(all_out_trans)
                    
                        for tran in dictMerged2.values():
                            src, dst = tran.src, tran.dst
                            if src is None and dst == ssid:  # it is a default transition
                                default_tran = tran
                            elif src == ssid:  # the src of tran is this state
                                out_trans.append(tran)
                            else:
                                all_out_trans[tran.ssid] = tran
                        out_trans.sort(key=operator.attrgetter("order"))

                        # Get inner_trans
                        inner_trans = list()
                        grandchildren = [grandchild for grandchild in child.childNodes if grandchild.nodeName == "Children"]
                        assert len(grandchildren) <= 1
                        if len(grandchildren) == 1:
                            inner_trans_dict = get_transitions(grandchildren[0].childNodes)
                            for tran in inner_trans_dict.values():
                                src, dst = tran.src, tran.dst
                                if src == ssid:
                                    inner_trans.append(tran)
                        inner_trans.sort(key=operator.attrgetter("order"))

                        # Create a state object
                        if state_type == "AND_STATE":
                            assert default_tran is None and out_trans == [], \
                                "Parse XML: AND_STATE should not have default transition or outgoing transitions."
                            order = int(get_attribute_value(child, "executionOrder"))
                            _state = AND_State(ssid=ssid, inner_trans=inner_trans, name=name, en=en, du=du, ex=ex,
                                               order=order)
                        elif state_type == "OR_STATE":
                            _state = OR_State(ssid=ssid, out_trans=out_trans, inner_trans=inner_trans, name=name,
                                            en=en, du=du, ex=ex, default_tran=default_tran)
                        else:
                            print(state_type)

                        # Get children states and junctions recursively
                        child_states, child_junctions, child_functions = get_children(child)
                        _state.funs = child_functions
                        for _child in child_states + child_junctions:
                            _child.father = _state
                            _state.children.append(_child)
                            if isinstance(_child, Junction) and _child.type == "HISTORY_JUNCTION":
                                if isinstance(_state, OR_State):
                                    _state.has_history_junc = True

                        if _state.children and isinstance(_state.children[0], AND_State):
                            _state.children.sort(key=operator.attrgetter("order"))
                        _states.append(_state)

                    else:
                        raise NotImplementedError("Unrecognized state type: %s" % state_type)
                    
                elif child.nodeName == "junction":
                    ssid = child.getAttribute("SSID")
                    junc_type = get_attribute_value(block=child, attribute="type")
                    # Get default_tran and out_trans
                    default_tran = None
                    out_trans = list()

                    dictMerged2 = dict(out_trans_dict)
                    dictMerged2.update(all_out_trans)

                    for tran in dictMerged2.values():
                        src, dst = tran.src, tran.dst
                        if src is None and dst == ssid:  # it is a default transition
                            default_tran = tran
                        elif src == ssid:  # the src of tran is this state
                            out_trans.append(tran)
                        else:
                            all_out_trans[tran.ssid] = tran
                    out_trans.sort(key=operator.attrgetter("order"))
                    # Create a junction object and put it into _junstions
                    _junctions.append(Junction(ssid=ssid, out_trans=out_trans,junc_type=junc_type, default_tran=default_tran))   
            return _states, _junctions, _functions

        # Create charts
        charts = self.model.getElementsByTagName(name="chart")
        for chart in charts:
            all_out_trans = dict()
            chart_id = chart.getAttribute("id")
            chart_name = replace_spaces(get_attribute_value(block=chart, attribute="name"))

            # A chart is wrapped into an AND-state
            chart_state = AND_State(ssid=chart_id, name=chart_name)
            states, junctions, functions = get_children(block=chart)
            # for fun in functions:
            #     if fun.type == "GRAPHICAL_FUNCTION":
            #         fun.chart_state1.father=chart_state
            #         chart_state.children.append(fun.chart_state1)
            chart_state.funs = functions
            for state in states + junctions:
                state.father = chart_state
                chart_state.children.append(state)
            if chart_state.children and isinstance(chart_state.children[0], AND_State):
                chart_state.children.sort(key=operator.attrgetter("order"))
            assert chart_state.check_children()
           
            chart_st = get_attribute_value(block=chart, attribute="sampleTime")
            if chart_st:
                if '.' in chart_st or 'e' in chart_st:
                    chart_st = Decimal(chart_st)
                else:
                    chart_st = int(chart_st)
            else:
                chart_st = -1

            chart_data = dict()
            message_dict = dict()
            event_dict = dict()
            if chart.getElementsByTagName(name="event"):
                for i, e in enumerate(chart.getElementsByTagName(name="event")):
                    event_name = get_attribute_value(e, "name")
                    event_scope = get_attribute_value(e, "scope")
                    port = i + 1
                    if get_attribute_value(e, "trigger"):
                        event_trigger = get_attribute_value(e, "trigger")
                    else:
                        event_trigger = None
                    event = SF_Event(name=event_name, scope=event_scope, trigger=event_trigger, port=port)
                    event_dict[event_name] = event

            for data in chart.getElementsByTagName(name="data"):
                var_name = data.getAttribute("name")
                value = get_attribute_value(data, "initialValue")
                if value is not None:
                    try:
                        value = expr_parser.parse(value)
                    except lark.exceptions.UnexpectedToken as e:
                        print("When parsing value:", value)
                        raise e
                else:
                    value = 0
                scope = get_attribute_value(data, "scope")
                sf_data = SF_Data(name=var_name, value=value, scope=scope)

                if data.getElementsByTagName(name="message"):
                    mesgNode = data.getElementsByTagName(name="message")[0]
                    if get_attribute_value(mesgNode, "isMessage") == "1":
                        message = SF_Message(name=var_name, data=0, scope=scope)
                        message_dict[var_name] = message
                else:
                    chart_data[var_name] = sf_data

            assert chart_name not in self.chart_parameters
            self.chart_parameters[chart_name] = {
                "state": chart_state,
                "data": chart_data,
                "st": chart_st,
                "message_dict": message_dict,
                "event_dict": event_dict
            }

    def parse_xml(self, default_SampleTimes=()):
        # Extract BlockParameterDefaults
        if not default_SampleTimes:
            default_SampleTimes = dict()
            default_para_blocks = self.model.getElementsByTagName("BlockParameterDefaults")
            assert len(default_para_blocks) == 1
            for child in default_para_blocks[0].childNodes:
                if child.nodeName == "Block":
                    child_type = child.getAttribute("BlockType")
                    assert child_type not in default_SampleTimes
                    default_SampleTimes[child_type] = get_attribute_value(child, "SampleTime")

        self.parse_stateflow_xml()

        # Extract name of the model
        models = self.model.getElementsByTagName("Model")
        assert len(models) <= 1
        self.name = ""
        if models:
            self.name = models[0].getAttribute("Name")

        system = self.model.getElementsByTagName("System")[0]

        # Add blocks
        blocks = [child for child in system.childNodes if child.nodeName == "Block"]

        # The following dictionary is used to replace port names as
        # formatted ones. The name of each port shoud be in the form of
        # port_type + port_number, such as in_0 and out_1 in order to delete
        # subsystems successfully (see function delete_subsystems in get_hcsp.py).
        port_name_dict = dict()  # mapping from old name to new name

        for block in blocks:
            # Type of block
            block_type = block.getAttribute("BlockType")

            # Block name.
            block_name = replace_spaces(block.getAttribute("Name"))

            # Sample time of block
            sample_time = get_attribute_value(block, "SampleTime")
            if (not sample_time) and block_type in default_SampleTimes:
                sample_time = default_SampleTimes[block_type]
            sample_time = parse_sample_time(sample_time)

            # For each type of block, obtain additional parameters
            if block_type == "Mux":
                inputs = get_attribute_value(block, "Inputs")
                ports = list(arg.value for arg in expr_parser.parse(get_attribute_value(block=block, attribute="Ports")).args)
                self.add_block(Mux(name=block_name, inputs=inputs, ports=ports))
            elif block_type == "DataStoreMemory":
                init_value = get_attribute_value(block=block, attribute="InitialValue")
                if init_value is not None:
                    try:
                        value = expr_parser.parse(init_value)
                    except lark.exceptions.UnexpectedToken as e:
                        print("When parsing value:", init_value)
                        raise e
                else:
                    value = 0
                dataStoreName = get_attribute_value(block, "DataStoreName")
                self.add_block(DataStoreMemory(name=block_name, value=value, dataStoreName=dataStoreName))
            elif block_type == "DataStoreRead":
                dataStoreName = get_attribute_value(block,"DataStoreName")
                self.add_block(DataStoreRead(name=block_name, dataStoreName=dataStoreName))
            elif block_type == "Constant":
                value = get_attribute_value(block, "Value")
                value = parse_value(value, default=1)
                self.add_block(Constant(name=block_name, value=value))
            elif block_type == "Integrator":
                init_value = get_attribute_value(block, "InitialCondition")
                init_value = parse_value(init_value, default=0)
                self.add_block(Integrator(name=block_name, init_value=init_value))
            elif block_type == "Logic":  # AND, OR, NOT
                op_name = get_attribute_value(block, "Operator")
                inputs = get_attribute_value(block, "Inputs")
                num_dest = int(inputs) if inputs else 2
                if op_name == "OR":
                    self.add_block(Or(name=block_name, num_dest=num_dest, st=sample_time))
                elif op_name == "NOT":
                    self.add_block(Not(name=block_name, st=sample_time))
                else:  # op_name == None, meaning it is an AND block
                    self.add_block(And(name=block_name, num_dest=num_dest, st=sample_time))
            elif block_type == "RelationalOperator":
                relation = get_attribute_value(block, "Operator")
                if relation == "~=":
                    relation = "!="
                self.add_block(Relation(name=block_name, relation=relation, st=sample_time))
            elif block_type == "Reference":
                # check if it is a discrete PID
                block_type = get_attribute_value(block, "SourceType")
                if block_type and block_type.startswith("PID"):
                    controller = get_attribute_value(block, "Controller")
                    st = eval(get_attribute_value(block, "SampleTime"))
                    assert get_attribute_value(block, "IntegratorMethod") == "Backward Euler"
                    assert get_attribute_value(block, "FilterMethod") == "Forward Euler"
                    assert get_attribute_value(block, "AntiWindupMode") == "back-calculation"
                    p = eval(get_attribute_value(block, "P"))
                    i = eval(get_attribute_value(block, "I"))
                    d = eval(get_attribute_value(block, "D"))
                    init_value = eval(get_attribute_value(block, "InitialConditionForIntegrator"))
                    upper_limit = eval(get_attribute_value(block, "UpperSaturationLimit"))
                    lower_limit = eval(get_attribute_value(block, "LowerSaturationLimit"))
                    kb = eval(get_attribute_value(block, "Kb"))
                    self.add_block(DiscretePID(name=block_name, controller=controller, st=st, pid=[p, i, d],
                                               init_value=init_value, saturation=[lower_limit, upper_limit], kb=kb))
                    continue

                relop = get_attribute_value(block, "relop")
                if relop:
                    if relop == "~=":
                        relop = "!="
                    self.add_block(Reference(name=block_name, relop=relop, st=sample_time))
                if block_type == "Digital clock":
                    period = get_attribute_value(block, "period")
                    self.chart_parameters['st']=float(period)/2

                    self.add_block(Clock(name=block_name,period=period))
            elif block_type == "Abs":
                self.add_block(Abs(name=block_name, st=sample_time))
            elif block_type == "Sqrt":
                sqrt_operator = get_attribute_value(block, "Operator")
                self.add_block(Sqrt(name=block_name, operator=sqrt_operator, st=sample_time))
            elif block_type == "Math":
                math_operator = get_attribute_value(block, "Operator")
                self.add_block(Square(name=block_name, operator=math_operator, st=sample_time))
            elif block_type == "Sum":
                inputs = get_attribute_value(block, "Inputs")
                dest_spec = inputs.replace("|", "") if inputs else "++"
                self.add_block(Add(name=block_name, dest_spec=dest_spec, st=sample_time))
            elif block_type == "Bias":
                bias = get_attribute_value(block, "Bias")
                bias = eval(bias) if bias else 0
                self.add_block(Bias(name=block_name, bias=bias, st=sample_time))
            elif block_type == "Product":
                inputs = get_attribute_value(block, "Inputs")
                dest_spec = inputs.replace("|", "") if inputs else "**"
                self.add_block(Product(name=block_name, dest_spec=dest_spec, st=sample_time))
            elif block_type == "Gain":
                factor = get_attribute_value(block, "Gain")
                factor = eval(factor) if factor else 1
                self.add_block(Gain(name=block_name, factor=factor, st=sample_time))
            elif block_type == "Sin":
                amplitude = parse_value(get_attribute_value(block, "Amplitude"), default=1)
                bias = parse_value(get_attribute_value(block, "Bias"), default=0)
                frequency = parse_value(get_attribute_value(block, "Frequency"), default=1)
                phase = parse_value(get_attribute_value(block, "Phase"), default=0)
                self.add_block(Sine(name=block_name, amplitude=amplitude, bias=bias, frequency=frequency, phase=phase, st=sample_time))
            elif block_type == "Saturate":
                upper_limit = parse_value(get_attribute_value(block, "UpperLimit"), default=0.5)
                lower_limit = parse_value(get_attribute_value(block, "LowerLimit"), default=-0.5)
                self.add_block(Saturation(name=block_name, up_lim=upper_limit, low_lim=lower_limit, st=sample_time))
            elif block_type == "UnitDelay":
                init_value = parse_value(get_attribute_value(block, "InitialCondition"), default=0)
                self.add_block(UnitDelay(name=block_name, init_value=init_value, st=sample_time))
            elif block_type == "MinMax":
                fun_name = get_attribute_value(block, "Function")
                fun_name = fun_name if fun_name else "min"
                input_num = get_attribute_value(block, "Inputs")
                input_num = int(input_num) if input_num else 1
                self.add_block(MinMax(name=block_name, num_dest=input_num, fun_name=fun_name, st=sample_time))
            elif block_type == "Switch":
                criteria = get_attribute_value(block, "Criteria")
                relation = ">" if criteria == "u2 > Threshold" else ("!=" if criteria == "u2 ~= 0" else ">=")
                threshold = get_attribute_value(block, "Threshold")
                threshold = eval(threshold) if threshold else 0
                self.add_block(Switch(name=block_name, relation=relation, threshold=threshold, st=sample_time))
            elif block_type == "DiscretePulseGenerator":
                pulseType = get_attribute_value(block, "PulseType") if get_attribute_value(block, "PulseType") else "Sample based"
                amplitude = float(get_attribute_value(block, "Amplitude")) if get_attribute_value(block, "Amplitude") else 1.0
                period = Decimal(get_attribute_value(block, "Period"))
                pulseWidth = Decimal(get_attribute_value(block, "PulseWidth"))
                phaseDelay = Decimal(get_attribute_value(block, "PhaseDelay")) if get_attribute_value(block, "PhaseDelay") else Decimal("0.0")
                self.add_block(DiscretePulseGenerator(
                    name=block_name, pulseType=pulseType, amplitude=amplitude, period=period,
                    pulseWidth=pulseWidth, phaseDelay=phaseDelay))
            elif block_type == "Inport":
                port_number = get_attribute_value(block, "Port")
                if not port_number:
                    port_number = "1"
                assert block_name not in port_name_dict
                port_name_dict[block_name] = "in_" + str(int(port_number) - 1)
                self.add_block(Port(name=port_name_dict[block_name], port_name=block_name, port_type="in_port"))
            elif block_type == "Outport":
                port_number = get_attribute_value(block=block, attribute="Port")
                if not port_number:
                    port_number = "1"
                assert block_name not in port_name_dict
                port_name_dict[block_name] = "out_" + str(int(port_number) - 1)
                self.add_block(Port(name=port_name_dict[block_name], port_name=block_name, port_type="out_port"))
            elif block_type == "Scope":
                num_dest = 0
                for child in system.childNodes:
                    if child.nodeName == "Line":
                        if get_attribute_value(block=child, attribute="DstBlock", name=block_name):
                            name = get_attribute_value(block=child, attribute="Name")
                            assert name is not None, "Scope output line is not named"
                            self.outputs.append(name)
                            num_dest += 1
                self.add_block(Scope(name=block_name, num_dest=num_dest, st=sample_time))
            elif block_type == "SubSystem":
                subsystem = block.getElementsByTagName("System")[0]

                # Check if it is a Signal Builder
                is_signal_builder = False
                for child in block.childNodes:
                    if child.nodeName == "Object" and child.getAttribute("PropName") == "MaskObject":
                        if get_attribute_value(block=child, attribute="Type") == "Sigbuilder block":
                            is_signal_builder = True
                            break
                if is_signal_builder:
                    signal_names = []
                    time_axises = []
                    data_axises = []
                    out_indexs = []
                    for node in subsystem.getElementsByTagName("Array"):
                        if node.getAttribute("PropName") == "Signals":
                            # assert node.getAttribute("Type") == "SigSuiteSignal"
                            for obj in node.getElementsByTagName("Object"):
                                signal_names.append(block_name + "_" + get_attribute_value(obj, "Name"))
                                xData = get_attribute_value(obj, "XData")
                                out_index = get_attribute_value(obj, "outIndex")
                                out_indexs.append(out_index)
                                # if "\n" in xData:
                                #     xData = xData.split("\n")[1]
                                assert xData.count('[') == xData.count(']') == 1
                                start = xData.index('[')
                                end = xData.index(']')
                                time_axises.append(tuple(parse_sample_time(e) for e in xData[start+1:end].split(',')))
                                yData = get_attribute_value(obj, "YData")
                                # if "\n" in yData:
                                #     yData = yData.split("\n")[1]
                                assert yData.count('[') == yData.count(']') == 1
                                start = yData.index('[')
                                end = yData.index(']')
                                data_axises.append(tuple(float(e) for e in yData[start+1:end].split(',')))

                    if not signal_names:
                        for node in subsystem.getElementsByTagName("Object"):
                            if node.getAttribute("PropName") == "Signals":
                                signal_names.append(block_name + "_" + get_attribute_value(node, "Name"))
                                xData = get_attribute_value(node, "XData")
                                # if "\n" in xData:
                                #     xData = xData.split("\n")[1]
                                # time_axises.append(tuple(float(e) for e in xData[1:-1].split(',')))
                                assert xData.count('[') == xData.count(']') == 1
                                start = xData.index('[')
                                end = xData.index(']')
                                time_axises.append(tuple(parse_sample_time(e) for e in xData[start + 1:end].split(',')))
                                yData = get_attribute_value(node, "YData")
                                # if "\n" in yData:
                                #     yData = yData.split("\n")[1]
                                # data_axises.append(tuple(float(e) for e in yData[1:-1].split(',')))
                                assert yData.count('[') == yData.count(']') == 1
                                start = yData.index('[')
                                end = yData.index(']')
                                data_axises.append(tuple(float(e) for e in yData[start + 1:end].split(',')))

                    assert signal_names
                    self.add_block(SignalBuilder(name=block_name, signal_names=tuple(signal_names),
                                                 time_axises=time_axises, data_axises=data_axises))
                    continue

                # Check if it is a stateflow chart
                sf_block_type = get_attribute_value(block, "SFBlockType")
                if sf_block_type == "Chart":
                    chart_paras = self.chart_parameters[block_name]
                    ports = list(arg.value for arg in expr_parser.parse(get_attribute_value(block=block, attribute="Ports")).args)
                    if len(ports) == 0:
                        ports = [0, 0]
                    elif len(ports) == 1:
                        ports.append(0)
                    num_dest, num_src = ports[:2]

                    # Check if it a triggered stateflow chart
                    triggers = [child for child in subsystem.childNodes if child.nodeName == "Block" and
                                child.getAttribute("BlockType") == "TriggerPort"]
                    assert len(triggers) <= 1
                    st = -2 if triggers else chart_paras['st']

                    stateflow = SF_Chart(name=block_name, state=chart_paras["state"], data=chart_paras["data"],
                                         num_src=num_src, num_dest=num_dest, st=st)

                    if triggers:  # it is a triggered stateflow chart
                        trigger_type = get_attribute_value(triggers[0], "TriggerType")
                        
                        if not trigger_type:
                            trigger_type = "rising"
                        assert trigger_type in ["rising", "falling", "either","function-call"]
                        stateflow.trigger_type = trigger_type
                        stateflow.num_dest += 1
                        stateflow.dest_lines.append(None)
                        # Extract the input events
                        stateflow_blocks = self.model.getElementsByTagName("Stateflow")
                        assert len(stateflow_blocks) == 1
                        charts = stateflow_blocks[0].getElementsByTagName("chart")
                        for chart in charts:
                            if get_attribute_value(block=chart, attribute="name", name=stateflow.name):
                                events = chart.getElementsByTagName("event")
                                for event in events:
                                    event_name = get_attribute_value(block=event, attribute="name")
                                    event_scope = get_attribute_value(block=event, attribute="scope")
                                    if event_scope == "INPUT_EVENT":
                                        event_trigger = get_attribute_value(block=event, attribute="trigger")
                                        if event_trigger == "RISING_EDGE_EVENT":
                                            event_trigger = "rising"
                                        elif event_trigger == "FALLING_EDGE_EVENT":
                                            event_trigger = "falling"
                                        elif event_trigger == "EITHER_EDGE_EVENT":
                                            event_trigger = "either"
                                        elif event_trigger == "FUNCTION_CALL_EVENT":
                                            event_trigger = "function-call"
                                        else:
                                            raise RuntimeError("Not implemented yet")
                                        stateflow.input_events.append((event_trigger, event_name))
                                break

                    assert stateflow.port_to_in_var == dict() and stateflow.port_to_out_var == dict()
                    for child in subsystem.childNodes:
                        if child.nodeName == "Block":
                            if child.getAttribute("BlockType") == "Inport":
                                in_var = child.getAttribute("Name")
                                port_id = get_attribute_value(child, "Port")
                                port_id = int(port_id) - 1 if port_id else 0
                                assert port_id not in stateflow.port_to_in_var
                                stateflow.port_to_in_var[port_id] = in_var
                            elif child.getAttribute("BlockType") == "Outport":
                                out_var = child.getAttribute("Name")
                                if out_var.find("()") == -1:
                                    port_id = get_attribute_value(child, "Port")
                                    port_id = int(port_id) - 1 if port_id else 0
                                    assert port_id not in stateflow.port_to_out_var
                                    stateflow.port_to_out_var[port_id] = out_var

                    self.add_block(stateflow)
                    continue

                # Check if it is a triggered or enabled subsystem
                triggers = [child for child in subsystem.childNodes if child.nodeName == "Block" and
                            child.getAttribute("BlockType") == "TriggerPort"]
                enables = [child for child in subsystem.childNodes if child.nodeName == "Block" and
                           child.getAttribute("BlockType") == "EnablePort"]
                # Enabled and triggered subsystems haven't been considered
                assert len(triggers) + len(enables) <= 1
                if triggers:  # it is a triggered subsystem
                    trigger_type = get_attribute_value(triggers[0], "TriggerType")
                    if not trigger_type:
                        trigger_type = "rising"
                    assert trigger_type in ["rising", "falling", "either", "function-call"]
                    ports = get_attribute_value(block, "Ports")
                    num_dest, num_src, _, _ = [int(port.strip("[ ]")) for port in ports.split(",")]
                    subsystem = Triggered_Subsystem(block_name, num_src, num_dest+1, trigger_type)
                    # Why num_src+1 above?
                    # A: The number of normal in-ports (num_src) + one TriggerPort
                elif enables:
                    assert get_attribute_value(enables[0], "StatesWhenEnabling") is None
                    assert get_attribute_value(enables[0], "PropagateVarSize") is None
                    ports = get_attribute_value(block, "Ports")
                    num_dest, num_src, _ = [int(port.strip("[ ]")) for port in ports.split(",")]
                    subsystem = Enabled_Subsystem(block_name, num_src, num_dest+1)
                else:  # it is a normal subsystem
                    ports = get_attribute_value(block=block, attribute="Ports")
                    num_dest, num_src = [int(port.strip("[ ]")) for port in ports.split(",")]
                    subsystem = Subsystem("subsystem", block_name, num_src, num_dest, st=sample_time)
                subsystem.diagram = SL_Diagram()
                # Parse subsystems recursively
                subsystem.diagram.model = block
                subsystem.diagram.parse_xml(default_SampleTimes)
                self.add_block(subsystem)

        # Add lines
        lines = [child for child in system.childNodes if child.nodeName == "Line"]
        for line in lines:
            line_name = get_attribute_value(block=line, attribute="Name")
            if not line_name:
                line_name = "?"
            ch_name = "?"
            src_block = replace_spaces(get_attribute_value(block=line, attribute="SrcBlock"))
            if src_block in port_name_dict:  # an input port
                ch_name = self.name + "_" + src_block
                src_block = port_name_dict[src_block]
            elif src_block not in self.blocks_dict:
                continue
            src_port = int(get_attribute_value(block=line, attribute="SrcPort")) - 1
            branches = [branch for branch in line.getElementsByTagName(name="Branch")
                        if not branch.getElementsByTagName(name="Branch")]
            if not branches:
                branches = [line]
            for branch in branches:
                dest_block = replace_spaces(get_attribute_value(block=branch, attribute="DstBlock"))
                if dest_block in port_name_dict:  # an output port
                    ch_name = self.name + "_" + dest_block
                    dest_block = port_name_dict[dest_block]
                dest_port = get_attribute_value(block=branch, attribute="DstPort")
                dest_port = -1 if dest_port in ["trigger", "enable"] else int(dest_port) - 1
                if dest_block in self.blocks_dict:
                    self.add_line(src=src_block, dest=dest_block, src_port=src_port, dest_port=dest_port,
                                  name=line_name, ch_name=ch_name)

        # The line name should keep consistent with the corresponding signals
        # if its src_block is a Signal Builder.
        for block in self.blocks_dict.values():
            if block.type == "signalBuilder":
                block.rename_src_lines()

    def add_block(self, block):
        """Add given block to the diagram."""
        assert block.name not in self.blocks_dict
        self.blocks_dict[block.name] = block

    def add_line(self, src, dest, src_port, dest_port, *, name="?", ch_name="?"):
        """Add given line to the diagram."""
        line = SL_Line(src, dest, src_port, dest_port, name=name, ch_name=ch_name)
        src_block = self.blocks_dict[line.src]
        dest_block = self.blocks_dict[line.dest]
        line.src_block = src_block
        line.dest_block = dest_block
        src_block.add_src(line.src_port, line)
        dest_block.add_dest(line.dest_port, line)

    def __str__(self):
        result = ""
        result += "\n".join(str(block) for block in self.blocks_dict.values())
        return result

    def check(self):
        """Checks the diagram is valid. Among the checks are:

        For each block, each dest port is filled, each src port has
        at least one outgoing line.

        """
        pass

    def add_line_name(self):
        """Give each group of lines a name."""

        # Set names of out-going lines from Signal Builders as the correspoding signals
        for block in self.blocks_dict.values():
            if isinstance(block, SignalBuilder):
                block.rename_src_lines()

        num_lines = 0
        for block in self.blocks_dict.values():
            # Give name to the group of lines containing each
            # incoming line (if no name is given already).
            for line in block.dest_lines:
                if line:
                    src, src_port = line.src, line.src_port
                    line_group = self.blocks_dict[src].src_lines[src_port]
                    if line_group[0].name == "?":
                        for line2 in line_group:
                            line2.name = "x" + str(num_lines)
                        num_lines += 1

            # Give name to each group of outgoing lines (if no
            # name is given already).
            for lines in block.src_lines:
                if lines and lines[0].name == "?":
                    for line in lines:
                        line.name = "x" + str(num_lines)
                    num_lines += 1

        # Add channel name for each line
        for block in self.blocks_dict.values():
            for line in block.dest_lines:
                if line:
                    assert line.name != "?"
                    if line.ch_name == "?":
                        line.ch_name = "ch_" + line.name + "_" + str(line.branch)
            for lines in block.src_lines:
                for line in lines:
                    assert line.name != "?"
                    if line.ch_name == "?":
                        line.ch_name = "ch_" + line.name + "_" + str(line.branch)

    def comp_inher_st(self):
        """Compute the sample time for each block with inherent sample time."""
        # Propagation
        terminate = False
        while not terminate:
            terminate = True
            for block in self.blocks_dict.values():
                if block.st == -1:

                    # Deal with triggered subsystems
                    if block.type == "triggered_subsystem":
                        trig_line = block.dest_lines[-1]
                        in_block = self.blocks_dict[trig_line.src]
                        if in_block.st >= 0:
                            block.st = in_block.st
                            block.is_continuous = (block.st == 0)
                            terminate = False
                        continue

                    in_st = []  # list of sample times of inputs of the block
                    for line in block.dest_lines:
                        in_block = self.blocks_dict[line.src]
                        # Xiong: Why can't in_block be a stateflow chart?
                        if not isinstance(in_block, SF_Chart) and in_block.st > 0:
                            in_st.append(in_block.st)
                        elif isinstance(in_block, Constant):
                            continue
                        else:
                            in_st = None
                            break
                    if in_st:
                        block.st = get_gcd(sample_times=in_st)
                        block.is_continuous = (block.st == 0)
                        terminate = False

        # Back-propagation
        terminate = False
        while not terminate:
            terminate = True
            for block in self.blocks_dict.values():
                if block.st == -1:
                    out_st = []  # list of sample times of outputs of the block
                    for lines in block.src_lines:
                        for line in lines:
                            out_block = self.blocks_dict[line.dest]
                            if out_block.st > 0:
                                out_st.append(out_block.st)
                            else:
                                out_st = []
                                break
                    if out_st:
                        block.st = get_gcd(sample_times=out_st)
                        block.is_continuous = (block.st == 0)
                        terminate = False
                
        # Re-classify the constant blocks
        for block in self.blocks_dict.values():
            if block.type == "constant":
                dest_block = self.blocks_dict[block.src_lines[0][0].dest]
                block.st = dest_block.st
                block.is_continuous = dest_block.is_continuous

    def inherit_to_continuous(self):
        for block in self.blocks_dict.values():
            if not isinstance(block, SF_Chart) and block.st == -1:
                block.st = 0
                block.is_continuous = True

    def delete_subsystems(self):
        """Unfold subsystems from the current diagram."""
        # Maintain list of subsystems (to be removed) and list of blocks
        # in those subsystems (to be added to self).
        subsystems = []
        blocks_in_subsystems = []

        for block in self.blocks_dict.values():
            if block.type in ["subsystem", "enabled_subsystem"]:
                # Collect all subsystems to be deleted
                subsystems.append(block.name)
                # The subsystem is treated as a diagram
                subsystem = block.diagram
                # Delete subsystems recursively
                subsystem.delete_subsystems()
                # Move all blocks except ports from the subsystem to the parent system
                for sub_block in subsystem.blocks_dict.values():
                    if sub_block.type not in ["in_port", "out_port"]:
                        blocks_in_subsystems.append(sub_block)
                # Sort the input ports in the subsystem by name
                input_ports = [sub_block for sub_block in subsystem.blocks_dict.values()
                               if sub_block.type == "in_port"]
                input_ports.sort(key=operator.attrgetter('name'))
                # Sort the output ports in the subsystem by name
                output_ports = [sub_block for sub_block in subsystem.blocks_dict.values()
                                if sub_block.type == "out_port"]
                output_ports.sort(key=operator.attrgetter('name'))

                # For each input line, find what is the source of this line
                # (in the current diagram or in other subsystems), and make the
                # new connections.
                for port_id in range(block.num_dest):
                    input_line = block.dest_lines[port_id]

                    # Find the source
                    src_block = None
                    if input_line.src in self.blocks_dict:
                        src_block = self.blocks_dict[input_line.src]
                    else:
                        for subsys in subsystems:
                            if input_line.src in self.blocks_dict[subsys].diagram.blocks_dict:
                                src_block = self.blocks_dict[subsys].diagram.blocks_dict[input_line.src]
                                break

                    # Delete the old line (input_line) from src_block
                    assert src_block is not None, "delete_subsystems: src_block not found."
                    src_block.src_lines[input_line.src_port][input_line.branch] = None

                    # Get the corresponding input port in the subsystem
                    port = input_ports[port_id]
                    assert port.name == "in_" + str(port_id)
                    for port_line in port.src_lines[0]:
                        dest_block = subsystem.blocks_dict[port_line.dest]
                        # Generate a new input line
                        new_input_line = SL_Line(src=src_block.name, dest=dest_block.name,
                                                 src_port=input_line.src_port, dest_port=port_line.dest_port)
                        new_input_line.name = input_line.name
                        # Delete the old line (port_line) and add a new one
                        dest_block.add_dest(port_id=port_line.dest_port, sl_line=new_input_line)
                        # Add a new line for src_block
                        src_block.add_src(port_id=input_line.src_port, sl_line=new_input_line)

                # For each output line, find what is the destination
                # (in the current diagram or in other diagrams), and make
                # the new connections.
                for port_id in range(block.num_src):
                    port = output_ports[port_id]

                    assert port.name == "out_" + str(port_id)
                    port_line = port.dest_lines[0]
                    src_block = subsystem.blocks_dict[port_line.src]

                    # Delete the old line (port_line) from src_block
                    src_block.src_lines[port_line.src_port][port_line.branch] = None
                    for output_line in block.src_lines[port_id]:
                        dest_block = None
                        if output_line.dest in self.blocks_dict:
                            dest_block = self.blocks_dict[output_line.dest]
                        else:
                            for subsys in subsystems:
                                if output_line.dest in self.blocks_dict[subsys].diagram.blocks_dict:
                                    dest_block = self.blocks_dict[subsys].diagram.blocks_dict[output_line.dest]
                                    break

                        # Generate a new output line
                        assert dest_block is not None, "delete_subsystems: dest_block not found"
                        new_output_line = SL_Line(src=src_block.name, dest=dest_block.name,
                                                  src_port=port_line.src_port, dest_port=output_line.dest_port)
                        new_output_line.name = output_line.name
                        # Delete the old line (output_line) and add a new one
                        dest_block.add_dest(port_id=output_line.dest_port, sl_line=new_output_line)
                        # Add a new line for src_block
                        src_block.add_src(port_id=port_line.src_port, sl_line=new_output_line)

        # Delete all the subsystems
        for name in subsystems:
            del self.blocks_dict[name]

        # Add new blocks from subsystems to block_dict
        for block in blocks_in_subsystems:
            assert block.name not in self.blocks_dict, "Repeated block name %s" % block.name
            self.blocks_dict[block.name] = block

    def separate_diagram(self):
        """Separate the diagram into the different parts, and stored in the
        corresponding variables in self.

        """
        for _, block in self.blocks_dict.items():
            # Continuous and discrete blocks contain the field is_continuous
            if hasattr(block, "is_continuous"):
                if isinstance(block, Scope):
                    self.scopes.append(block)
                elif block.is_continuous:
                    self.continuous_blocks.append(block)
                else:
                    self.discrete_blocks.append(block)
            elif isinstance(block, DataStoreMemory):
                self.dsms.append(block)
            else:
                assert False, "block: %s" % type(block)

    def seperate_diagram(self):
        """Seperate a diagram into discrete and continuous subdiagrams."""
        # delete in and out-ports
        blocks_dict = {name: block for name, block in self.blocks_dict.items()
                       if block.type not in ['in_port', 'out_port']}
        # Get stateflow charts and then delete them from block_dict
        sf_charts = [block for block in blocks_dict.values() if block.type == "stateflow"]
        for name in [block.name for block in sf_charts]:
            del blocks_dict[name]

        # Data store memory
        dataStoreMemorys = [block for block in blocks_dict.values() if block.type == "DataStoreMemory"]
        for name in [block.name for block in dataStoreMemorys]:
            del blocks_dict[name]

        # Data store read
        dataStoreReads = [block for block in blocks_dict.values() if block.type == "DataStoreRead"]
        for name in [block.name for block in dataStoreReads]:
            del blocks_dict[name]

        clocks=[block for block in blocks_dict.values() if block.type == "clock"]
        for name in [block.name for block in clocks]:
            del blocks_dict[name]

        muxs = [block for block in blocks_dict.values() if block.type == "mux"]
        for name in [block.name for block in muxs]:
            del blocks_dict[name]
        # Get buffers and then delete them from block_dict
        buffers = [block for block in blocks_dict.values() if block.type == "discrete_buffer"]
        for name in [block.name for block in buffers]:
            del blocks_dict[name]
        discretePulseGenerator=[block for block in blocks_dict.values() if block.type == "DiscretePulseGenerator"]
        for name in [block.name for block in discretePulseGenerator]:
            del blocks_dict[name]  
        # Seperating: search SCC by BFS
        discrete_subdiagrams = []
        continuous_subdiagrams = []
        while blocks_dict:
            _, block = blocks_dict.popitem()
            scc = [block]
            bs = [block]
            while bs:
                b = bs.pop(-1)
                for src_line in b.src_lines:
                    for line in src_line:
                        dest_name = line.dest
                        if dest_name in blocks_dict and blocks_dict[dest_name].is_continuous == block.is_continuous:
                            bs.append(blocks_dict[dest_name])
                            scc.append(blocks_dict[dest_name])
                            del blocks_dict[dest_name]
                for dest_line in b.dest_lines:
                    if dest_line !=None:
                        src_name = dest_line.src
                        if src_name in blocks_dict and blocks_dict[src_name].is_continuous == block.is_continuous:
                            bs.append(blocks_dict[src_name])
                            scc.append(blocks_dict[src_name])
                            del blocks_dict[src_name]
            if block.is_continuous:
                continuous_subdiagrams.append(scc)
            else:
                discrete_subdiagrams.append(scc)

        # Sort discrete blocks
        # discrete_subdiagrams_sorted = []
        # for scc in discrete_subdiagrams:
        #     scc_dict = {block.name: block for block in scc}
        #     sorted_scc = []
        #     while scc_dict:
        #         delete_blocks = []
        #         for block in scc_dict.values():
        #             src_set = set()
        #             for dest_line in block.dest_lines:
        #                 if dest_line !=None:
        #                     src_set.add(dest_line.src)
        #             if src_set.isdisjoint(set(scc_dict.keys())):
        #                 sorted_scc.append(block)
        #                 delete_blocks.append(block.name)
        #         for block_name in delete_blocks:
        #             del scc_dict[block_name]
        #     discrete_subdiagrams_sorted.append(sorted_scc)

        return discrete_subdiagrams, continuous_subdiagrams, sf_charts, buffers, \
            discretePulseGenerator, muxs, dataStoreMemorys, dataStoreReads, clocks

    def add_buffers(self):
        buffers = []
        for block in self.blocks_dict.values():
            if block.type in {"stateflow", "unit_delay"}:
                for port_id in range(len(block.dest_lines)):
                    line = block.dest_lines[port_id]
                    src_block = self.blocks_dict[line.src]
                    if not src_block.is_continuous:
                        if src_block.st == block.st:
                            continue
                        buffer = Discrete_Buffer(in_st=src_block.st, out_st=block.st)
                        buffers.append(buffer)
                        # Link src_block to buffer
                        line.dest = buffer.name
                        line.dest_port = 0
                        assert buffer.dest_lines == [None]
                        buffer.dest_lines = [line]
                        # Link buffer to block
                        line = SL_Line(src=buffer.name, dest=block.name, src_port=0, dest_port=port_id)
                        line.branch = 0
                        assert buffer.src_lines == [[]]
                        buffer.src_lines = [[line]]
                        block.dest_lines[port_id] = line
                
                for branch_list in block.src_lines:
                    for branch in branch_list:
                        dest_block = self.blocks_dict[branch.dest]
                        if not dest_block.is_continuous:
                            if block.st == dest_block.st:
                                continue
                            buffer = Discrete_Buffer(in_st=block.st, out_st=dest_block.st)
                            buffers.append(buffer)
                            # Link buffer to dest_block
                            line = SL_Line(src=buffer.name, dest=dest_block.name,
                                           src_port=0, dest_port=branch.dest_port)
                            line.branch = 0
                            assert buffer.src_lines == [[]]
                            buffer.src_lines = [[line]]
                            dest_block.dest_lines[line.dest_port] = line
                            # Link block to buffer
                            branch.dest = buffer.name
                            branch.dest_port = 0
                            assert buffer.dest_lines == [None]
                            buffer.dest_lines = [branch]
        for buffer in buffers:
            self.add_block(buffer)
        # Reset buffer number to 0
        Discrete_Buffer.num = 0



from geventwebsocket import WebSocketServer, WebSocketApplication, Resource
import gevent
from collections import OrderedDict
import sys
import traceback
import re
from os import listdir
from os.path import isfile, join, dirname, basename
from multiprocessing import Process, Queue
from queue import Empty
import signal
import sys



from operator import pos

from ss2hcsp.hcsp import expr, hcsp
from ss2hcsp.hcsp.parser import parse_hoare_triple_with_meta, parse_expr_with_meta
from ss2hcsp.hcsp.simulator import get_pos
from hhlpy.hhlpy_without_dG import CmdVerifier
from hhlpy.wolframengine_wrapper import wl_prove
from hhlpy.z3wrapper import z3_prove
from hhlpy.wolframengine_wrapper import session, found_wolfram


import json

def runCompute(code):
    """Compute the verification condition information by the code received from the client editor.
    Return an array of verification condition information."""
    hoare_triple = parse_hoare_triple_with_meta(code)

    # Initialize the verifier
    verifier = CmdVerifier(
        pre=hoare_triple.pre, 
        hp=hoare_triple.hp,
        post=hoare_triple.post,
        functions=hoare_triple.functions)

    # Compute wp and verify
    verifier.compute_wp()

    # Return verification condition informations
    vc_infos = []

    for _, vcs in verifier.get_all_vcs().items():

        for vc in vcs:
            assume = split_imp(vc.expr)[:-1]
            show = split_imp(vc.expr)[-1] 


            # Get the proof method
            label_computed = vc.comp_label
            method_stored = None
            
            pms_start_pos = -1
            pms_end_pos = -1

            # Use the bottom-most assertion to attach the VC to
            assert vc.bottom_loc is not None
            bottom_loc = vc.bottom_loc
 
            # Case when the bottom-most assertion of vc is a post condition.
            if bottom_loc.isPost:
                bottom_ast = hoare_triple.post[vc.bottom_loc.index]

            # Case when the bottom-most assertion of vc is a loop or an ode invariant.
            elif bottom_loc.isInv:
                assert bottom_loc.hp_pos is not None
                hp = get_pos(hoare_triple.hp, vc.bottom_loc.hp_pos)

                assert isinstance(hp, (hcsp.Loop, hcsp.ODE))
                bottom_ast = hp.inv[vc.bottom_loc.index]
            
            else:
                raise NotImplementedError

            proof_methods = bottom_ast.proof_methods
            pms_meta = proof_methods.meta
            if pms_meta.empty:
                pms_meta.start_pos = bottom_ast.meta.end_pos
                pms_meta.end_pos = bottom_ast.meta.end_pos
            pms_start_pos = proof_methods.meta.start_pos
            pms_end_pos = proof_methods.meta.end_pos
            # If the method of this vc is stored in proof_methods,
            # send the method back to the client, which will be used as the default method.
            for proof_method in proof_methods.pms:
                if str(label_computed) == str(proof_method.label):
                    method_stored = proof_method.method


            # Map origin positions in syntax tree to positions on the character level
            originRanges = []
            # Append the hcsp program ranges.
            for originPos in vc.path:
                hp = get_pos(hoare_triple.hp, originPos[0])

                if isinstance(hp, hcsp.ODE):
                    from_pos = hp.meta.start_pos + 1
                    to_pos = hp.constraint.meta.end_pos
                else:
                    from_pos = hp.meta.start_pos
                    to_pos = hp.meta.end_pos
                
                originRanges.append({"from": from_pos, "to": to_pos})

            # Append the expression ranges.
            for origin in vc.origins:
                print("origin:", origin)
                if origin.isPre:
                    originMeta = hoare_triple.pre[origin.index].meta
                    to = originMeta.end_pos
               
                elif origin.isConstraint:
                    assert origin.hp_pos is not None
                    hp = get_pos(hoare_triple.hp, origin.hp_pos)
                    assert isinstance(hp, hcsp.ODE)
                    originMeta = hp.constraint.meta
                    to = originMeta.end_pos

                elif origin.isInv or origin.isPost:
                    if origin.isInv:
                        assert origin.hp_pos is not None
                        hp = get_pos(hoare_triple.hp, origin.hp_pos)
                        assert isinstance(hp, (hcsp.Loop, hcsp.ODE))
                        assertion = hp.inv[origin.index]
                    else:
                        assertion = hoare_triple.post[origin.index]

                    originMeta = assertion.meta
                    proof_methods = assertion.proof_methods
                    pms_meta = proof_methods.meta
                    if pms_meta.empty:
                        pms_meta.start_pos = assertion.meta.end_pos
                        pms_meta.end_pos = assertion.meta.end_pos
                    
                    to = pms_meta.start_pos
                else:
                    raise NotImplementedError
                originRanges.append({"from": originMeta.start_pos, "to": to})


            vc_infos.append({
                "formula": str(vc.expr),
                "assume": [str(asm) for asm in assume],
                "show": str(show),
                "label": str(vc.comp_label),
                "method": method_stored,
                "origin": originRanges,
                "pms_start_pos": pms_start_pos,
                "pms_end_pos": pms_end_pos
            })
    
    return vc_infos

def runVerify(formula, code, solver):
    formulaExpr = parse_expr_with_meta(formula)
    hoare_triple = parse_hoare_triple_with_meta(code)

    if solver == "z3":
        return z3_prove(formulaExpr, functions=hoare_triple.functions)
    elif solver == "wolfram":
        return wl_prove(formulaExpr)
    else:
        raise NotImplementedError

# This function runs `runCompute` and `runVerify` on a separate process
# ensuring that the main process is still available to receive client requests
def runComputationProcess(inputQueue, outputQueue):
    if found_wolfram:
        print("Starting Wolfram Engine", flush=True)
        session.start_future()
        print("Started Wolfram Engine", flush=True)
        def handler(signum, frame):
            print("Stopping Wolfram Engine", flush=True)
            session.terminate()
            print("Process ended.", flush=True)
            sys.exit()
        signal.signal(signal.SIGTERM, handler)
    else:
        print("Wolfram Engine not found", flush=True)
    
    outputQueue.put({"type": "computation_process_ready"})

    while True:
        msg = inputQueue.get() # Wait for next verification task

        if msg["type"] == "compute":
            try:
                vcs = runCompute(code=msg["code"])
                outputQueue.put({"vcs": vcs, "type": "computed", "file": msg["file"]})
            except Exception as e:
                outputQueue.put({"error": str(e), "type": "computed", "file": msg["file"]})

        elif msg["type"] == "verify":
            result = runVerify(
                formula=msg["formula"],
                code=msg["code"],
                solver=msg["solver"])
        
            outputQueue.put({
                "file": msg["file"],
                "index": msg["index"], 
                "formula": msg["formula"], 
                "result": result, 
                "type": "verified"})

# These queues are used to communicate between the two processes
computationInputQueue = Queue()
computationOutputQueue = Queue()

def natural_sort(l): 
    convert = lambda text: int(text) if text.isdigit() else text.lower() 
    alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)] 
    return sorted(l, key=alphanum_key)

def getExampleList():
    path = join(dirname(__file__), "..", "examples")
    filenames = [f for f in listdir(path) if isfile(join(path, f))]
    filenames = natural_sort(filenames)
    return filenames

def getFileList(path):
    path = join(dirname(__file__), "..", "examples", path)
    filenames = [f for f in listdir(path) if isfile(join(path, f))]
    dirnames = [f for f in listdir(path) if not isfile(join(path, f))]
    filenames = natural_sort(filenames)
    dirnames = natural_sort(dirnames)
    return (dirnames, filenames)

def getFileCode(example):
    file = join(dirname(__file__), "../examples", example)
    file = open(file, mode='r', encoding='utf-8')
    code = file.read()
    file.close()
    return code

def split_imp(e):
    """Split an implication expression into a list of its sub-expressions in order.
    Example, p -> q -> u into [p, q, u]"""
    if isinstance(e, expr.LogicExpr) and e.op == '->':
        return split_imp(e.exprs[0]) + split_imp(e.exprs[1])
    else:
        return [e]

computationProcess = None
ws = None

class HHLPyApplication(WebSocketApplication):
    def on_open(self):
        print("Connection opened")
        global ws # TODO: Is there a better way than using global here?
        ws = self.ws

    def on_message(self, message):
        try:
            if message != None:
                print(message, flush=True)
                msg = json.loads(message)
                # If the type of message received is "compute", 
                # the message has a code field, which is the document in editor,
                # then call runCompute function.
                if msg["type"] == "compute":
                    computationInputQueue.put(msg)

                # If the type of message received is "verify",
                # the message has the index, the formula and solver of corresponding vc.
                elif msg["type"] == "verify":
                    computationInputQueue.put(msg)

                elif msg["type"] == "get_example_list":
                    examples = getExampleList()
                    result = {"examples": examples, "type": "example_list"}
                    self.ws.send(json.dumps(result)) 
                
                elif msg["type"] == "get_file_list":
                    (dirs, files) = getFileList(msg["path"])
                    result = {"dirs": dirs, "files": files, "path": msg["path"], "type": "file_list"}
                    self.ws.send(json.dumps(result)) 

                elif msg["type"] == "load_file":
                    code = getFileCode(msg["file"])
                    result = {"file": msg["file"], "code": code, "type": "load_file"}
                    self.ws.send(json.dumps(result)) 

                elif msg["type"] == "cancel_computation":
                    global computationProcess
                    computationProcess.terminate()
                    computationProcess.join()
                    computationProcess = startComputationProcess()

                elif msg["type"] == "save_file":
                    file = join(dirname(__file__), "../examples", msg["file"])
                    file = open(file, mode='w', encoding='utf-8')
                    file.write(msg["code"])
                    file.close()

                else:
                    raise NotImplementedError    
        except Exception as e:
            print(str(e), file=sys.stderr)
            print(traceback.format_exc(), file=sys.stderr)
            result_dict = {"error": str(e), "type": "error"}  
            self.ws.send(json.dumps(result_dict)) 

    def on_close(self, reason):
        print(reason)

    def on_error(self, ):
        print("Server Error")

def checkComputationProcess():
    if ws is not None:
        try:
            result = computationOutputQueue.get(False)
            print("out!", flush=True)
            ws.send(json.dumps(result))
        except Empty:
            pass

def startComputationProcess():
    computationProcess = Process(daemon=True, target=runComputationProcess, 
            kwargs={"inputQueue": computationInputQueue, "outputQueue": computationOutputQueue})
    computationProcess.start()
    return computationProcess

from flask import Flask, send_from_directory, redirect

def startHttpServer(port):
    print(dirname(__file__))
    app = Flask('hhlpy', root_path=dirname(__file__))

    @app.route('/')
    def index():
        return redirect('/index.html')

    @app.route('/<path:path>')
    def send_static(path):
        return send_from_directory('dist', path)

    app.run(port=port)

def startHttpProcess(port):
    httpProcess = Process(daemon=True, target=startHttpServer, args=[port])
    httpProcess.start()
    
if __name__ == "__main__":
    port = 8000
    websocketServer = WebSocketServer(
        ('', port),
        Resource(OrderedDict([('/', HHLPyApplication)]))
    )
    try:
        computationProcess = startComputationProcess()
        startHttpProcess(port)
        print("Running python websocket server on ws://localhost:{0}".format(port), flush=True)
        websocketServer.start()
        while True:
            checkComputationProcess() # Check whether there are new computation results
            gevent.sleep(0) # Pause to allow the server to handle new requests from the client
    except KeyboardInterrupt:
        print("KeyboardInterrupt")
        print("Closing python websocket server")
        websocketServer.stop()



from ss2hcsp.sl.sl_block import SL_Block
from ss2hcsp.hcsp.expr import AVar, AConst, true_expr, RelExpr, OpExpr
import ss2hcsp.hcsp.hcsp as hp


class Gain(SL_Block):
    """Multiply dest line by a factor."""
    def __init__(self, name, factor=1, st=-1):
        super(Gain, self).__init__("gain", name, 1, 1, st)

        assert isinstance(factor, (int, float))
        self.factor = factor

    def get_expr(self):
        """Compute the assignment corresponding to gain block."""
        in_var = AVar(self.dest_lines[0].name)
        return OpExpr("*", in_var, AConst(self.factor))

    def __str__(self):
        expr = self.get_expr()
        out_var = self.src_lines[0][0].name
        return "%s: %s = %s  (st = %s)" % (self.name, out_var, str(expr), str(self.st))

    def __repr__(self):
        return "Gain(%s, %s, %s, in = %s, out = %s)" % \
            (self.name, self.factor, self.st, str(self.dest_lines), str(self.src_lines))

    def get_output_hp(self):
        expr = self.get_expr()
        out_var = self.src_lines[0][0].name
        return hp.Assign(out_var, expr)

    def get_var_subst(self):
        expr = self.get_expr()
        out_var = self.src_lines[0][0].name
        return {out_var: expr}

    def get_var_map(self):
        expr = self.get_expr()
        out_var = self.src_lines[0][0].name
        return {out_var: [(true_expr, expr)]}



from ss2hcsp.sl.sl_block import SL_Block
from ss2hcsp.hcsp import hcsp as hp
from ss2hcsp.hcsp.expr import AVar, AConst
from math import gcd


def lcm(x, y):
    assert isinstance(x, int) and x > 0 and isinstance(y, int) and y > 0
    return x * y // gcd(x, y)


class Discrete_Buffer(SL_Block):
    num = 0

    def __init__(self, in_st, out_st, name=None):
        super(Discrete_Buffer, self).__init__()
        self.type = "discrete_buffer"
        if not name:
            self.name = "buffer" + str(Discrete_Buffer.num)
            Discrete_Buffer.num += 1
        assert isinstance(in_st, int) and in_st > 0
        self.in_st = in_st
        assert isinstance(out_st, int) and out_st > 0
        self.out_st = out_st

        self.num_src = 1
        self.num_dest = 1

        self.src_lines = [[]]
        self.dest_lines = [None]

    def __str__(self):
        return "%s: Discrete_Buffer[in = %s, out = %s, in_st = %s, out_st = %s]" % \
               (self.name, str(self.dest_lines), str(self.src_lines), str(self.in_st), str(self.out_st))

    def get_hcsp(self):
        assert len(self.dest_lines) == 1
        in_var = self.dest_lines[0].name
        branch = self.dest_lines[0].branch
        in_channel = hp.InputChannel(ch_name="ch_" + in_var + "_" + str(branch), var_name=in_var)

        assert len(self.src_lines) == 1 and len(self.src_lines[0]) == 1
        out_var = self.src_lines[0][0].name
        branch = self.src_lines[0][0].branch
        out_channel = hp.OutputChannel(ch_name="ch_" + out_var + "_" + str(branch), expr=AVar(in_var))

        channel_seq = []
        for i in range(lcm(self.in_st, self.out_st)):
            if i % self.in_st == 0 and i % self.out_st == 0:
                channel_seq.append(hp.Sequence(in_channel, out_channel))
            elif i % self.in_st == 0:
                channel_seq.append(in_channel)
            elif i % self.out_st == 0:
                channel_seq.append(out_channel)
            else:
                channel_seq.append(None)

        # Add delays into channel_seq
        channel_delay_seq = []
        for e in channel_seq:
            if isinstance(e, hp.HCSP):
                channel_delay_seq.append(e)
            channel_delay_seq.append("delay")

        hcsp_queue = []
        i = 0
        while i < len(channel_delay_seq):
            if isinstance(channel_delay_seq[i], hp.HCSP):
                hcsp_queue.append(channel_delay_seq[i])
                i += 1
            else:  # channel_delay_seq[i] == "delay"
                delay = 0
                while i < len(channel_delay_seq) and not isinstance(channel_delay_seq[i], hp.HCSP):
                    assert channel_delay_seq[i] == "delay"
                    delay += 1
                    i += 1
                hcsp_queue.append(hp.Wait(AConst(delay)))

        return hp.Loop(hp.Sequence(*hcsp_queue))



import matplotlib
import matplotlib as plt
matplotlib.use("TkAgg")
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
from matplotlib.figure import Figure

import tkinter as tk
from tkinter import ttk

import numpy as np 

LARGE_FONT= ("Verdana", 12)

# Change the res dict here with what we got from exec_parallel
# function. 
res = { 
    'time_series': {
    'P1': 0,
    'P2': 1
    }
}

def get_program(res):
    return res.get('time_series').keys()

# def get_total_time(res):
#     return res.get('time') 

def get_data_time(res, ProgramName):
    DataState = {}
    temp = res.get("time_series")
    lst = temp.get(ProgramName)
    for t in lst:
        state = t.get("state")
        # print(t.get('time'))
        for key in state.keys():
            # if state.get(key) == {}:
            #     pass
            if key not in DataState.keys():
                DataState.update({key:([],[])})
            # print (DataState.get(key))
            # print (state.get())
            DataState.get(key)[0].append(state.get(key))
            DataState.get(key)[1].append(t.get('time'))
    return DataState



class Graphapp(tk.Tk):

    def __init__(self, res, *args, **kwargs):
        
        tk.Tk.__init__(self, *args, **kwargs)

        #tk.Tk.iconbitmap(self, default="clienticon.ico")
        tk.Tk.wm_title(self, "Graph GUI")
        
        
        container = tk.Frame(self)
        container.pack(side="top", fill="both", expand = True)
        container.grid_rowconfigure(0, weight=1)
        container.grid_columnconfigure(0, weight=1)
        self.res = res
        self.frames = {}

        for F in (StartPage, PageOne, PageTwo):

            frame = F(container, self)

            self.frames[F] = frame

            frame.grid(row=0, column=0, sticky="nsew")

        frame = PageThree(container, self, self.res)

        self.frames[PageThree] = frame

        frame.grid(row=0, column=0, sticky="nsew")
        
        self.show_frame(StartPage)

    def show_frame(self, cont):

        frame = self.frames[cont]
        frame.tkraise()

        
class StartPage(tk.Frame):

    def __init__(self, parent, controller):
        tk.Frame.__init__(self,parent)
        label = tk.Label(self, text="Start Page", font=LARGE_FONT)
        label.pack(pady=10,padx=10)

        button1 = ttk.Button(self, text="Visit Graphing",
                            command=lambda: controller.show_frame(PageThree))
        button1.pack()

        button2 = ttk.Button(self, text="Quit",
                            command=lambda: quit)
        button2.pack()



class PageOne(tk.Frame):

    def __init__(self, parent, controller):
        tk.Frame.__init__(self, parent)
        label = tk.Label(self, text="Page One!!!", font=LARGE_FONT)
        label.pack(pady=10,padx=10)

        button1 = ttk.Button(self, text="Back to Home",
                            command=lambda: controller.show_frame(StartPage))
        button1.pack()

        button2 = ttk.Button(self, text="Page Two",
                            command=lambda: controller.show_frame(PageTwo))
        button2.pack()


class PageTwo(tk.Frame):

    def __init__(self, parent, controller):
        tk.Frame.__init__(self, parent)
        label = tk.Label(self, text="Page Two!!!", font=LARGE_FONT)
        label.pack(pady=10,padx=10)

        button1 = ttk.Button(self, text="Back to Home",
                            command=lambda: controller.show_frame(StartPage))
        button1.pack()

        button2 = ttk.Button(self, text="Page One",
                            command=lambda: controller.show_frame(PageOne))
        button2.pack()


class PageThree(tk.Frame):

    def __init__(self, parent, controller, res):
        tk.Frame.__init__(self, parent)
        label = tk.Label(self, text="Graph Page!", font=LARGE_FONT)
        label.pack(pady=10,padx=10)
        self.res = res
        button1 = ttk.Button(self, text="Back to Home",
                            command=lambda: controller.show_frame(StartPage))
        button1.pack()

        label1 = tk.Label(self, text="Program")
        label1.pack()
        e1 = tk.Entry(self, bd = 5)
        e1.pack()

        f = Figure(figsize=(5,5), dpi=100)
        a = f.add_subplot(111)
        

        def plot_graph():
            if (e1.get() in get_program(self.res)):
                # x = np.arange(0,4*np.pi,0.1)   # start,stop,step
                # y1 = np.sin(x)
                # y2 = np.cos(x)
                # a.plot(x,y1)
                # a.plot(x,y2) 
                dataset_state = get_data_time(self.res, e1.get())
                for t in dataset_state.keys():
                    x = dataset_state.get(t)[1]
                    y = dataset_state.get(t)[0]
                    a.plot(x,y,label = t)
                    a.legend()
                canvas = FigureCanvasTkAgg(f, self)
                canvas.draw()
                canvas.get_tk_widget().pack(side=tk.BOTTOM, fill=tk.BOTH, expand=True)

                toolbar = NavigationToolbar2Tk(canvas, self)
                toolbar.update()
                canvas._tkcanvas.pack(side=tk.TOP, fill=tk.BOTH, expand=True)


        button2 = ttk.Button(self, text="Draw",
                            command=plot_graph)
        button2.pack()

        # canvas = FigureCanvasTkAgg(f, self)
        # canvas.draw()
        # canvas.get_tk_widget().pack(side=tk.BOTTOM, fill=tk.BOTH, expand=True)

        # toolbar = NavigationToolbar2Tk(canvas, self)
        # toolbar.update()
        # canvas._tkcanvas.pack(side=tk.TOP, fill=tk.BOTH, expand=True)


def graph(res, proc_name, tkplot=False, separate=True, variables=None):
    DataState = {}
    temp = res.get("time_series")
    lst = temp.get(proc_name)
    for t in lst:
        state = t.get("state")
        for key in state.keys():
            if variables is not None and key not in variables:
                continue
            if key not in DataState.keys():
                DataState.update({key:([],[])})
            DataState.get(key)[0].append(state.get(key))
            DataState.get(key)[1].append(t.get('time'))
                
    if tkplot:
        app = Graphapp(res)
        app.mainloop()
    else:
        if separate:
            for t in DataState.keys():
                x = DataState.get(t)[1]
                y = DataState.get(t)[0]
                plt.plot(x, y, label=t)
                plt.show()
        else:
            for t in DataState.keys():
                x = DataState.get(t)[1]
                y = DataState.get(t)[0]
                plt.plot(x, y, label=t)
                plt.legend()



from decimal import Decimal

from ss2hcsp.sl.sl_block import SL_Block, get_gcd
from ss2hcsp.hcsp import hcsp as hp
from ss2hcsp.hcsp.expr import conj, AVar, AConst, OpExpr, RelExpr


class SignalBuilder(SL_Block):
    """Block for Signal Builder."""
    def __init__(self, name, signal_names=(), time_axises=(), data_axises=(), out_index=()):
        # Both time_axises and data_axises are lists of tuples of values, where
        # each tuple represents one series.
        assert len(time_axises) == len(data_axises) == len(signal_names)
        self.num_series = len(signal_names)
        for i in range(self.num_series):
            assert len(time_axises[i]) == len(data_axises[i])
            assert all(isinstance(v, Decimal) for v in time_axises[i])

        self.signal_names = [name.replace(' ', '_') for name in signal_names]
        self.time_axises = time_axises
        self.data_axises = data_axises
        self.out_indexs = out_index

        # Now compute the sample time
        st = get_gcd([t for t in self.time_axises[0] if t > 0])
        for axes in self.time_axises[1:]:
            st = get_gcd([st, *(t for t in axes if t > 0)])
        super(SignalBuilder, self).__init__("signalBuilder", name, len(signal_names), 0, st)

        # Variable representing tick within the signal builder
        self.tick_var = self.name + "_tick"

    def __str__(self):
        return "%s: SignalBuilder[signals = %s, out = %s]" % (self.name, str(self.signal_names), str(self.src_lines))

    def __repr__(self):
        return str(self)

    def get_init_hp(self):
        return hp.Assign(self.tick_var, AConst(0))

    def get_output_hp(self):
        procs = []
        for s_id in range(self.num_series):
            branches = []
            time_axis = self.time_axises[s_id]
            data_axis = self.data_axises[s_id]
            signal_name = self.signal_names[s_id]
            for i in range(len(time_axis)-1):
                left = int(time_axis[i] / self.st)
                right = int(time_axis[i+1] / self.st)
                if left == right:
                    continue
                cond = conj(RelExpr(">=", AVar(self.tick_var), AConst(left)),
                            RelExpr("<", AVar(self.tick_var), AConst(right)))
                slope = (data_axis[i+1] - data_axis[i]) / (right - left)
                act = hp.Assign(AVar(signal_name),
                    OpExpr("+", OpExpr("*", AConst(slope), OpExpr("-", AVar(self.tick_var), AConst(left))),
                                AConst(data_axis[i])))
                branches.append((cond, act))

            else_act = hp.Assign(AVar(signal_name), AConst(data_axis[-1]))
            procs.append(hp.ITE(branches, else_act))
        procs.append(hp.Assign(self.tick_var, OpExpr("+", AVar(self.tick_var), AConst(1))))
        return hp.seq(procs)

    def rename_src_lines(self):
        assert len(self.signal_names) == len(self.src_lines)
        for i in range(len(self.signal_names)):
            for line in self.src_lines[i]:
                line.name = self.signal_names[i]



from ss2hcsp.sl.sl_block import SL_Block
from ss2hcsp.hcsp import hcsp 
from ss2hcsp.hcsp.expr import AConst

class Clock(SL_Block):
    def __init__(self, name, period):
        # Clock has one output line and no input lines.
        super(Clock, self).__init__("clock", name, 1, 0, 0)
        self.period = period

    def __str__(self):
        return "%s: Clock(period = %s)" % (self.name,self.max_step)

    def __repr__(self):
        return str(self)

    def get_hcsp(self):
        return hcsp.Sequence(hcsp.Loop(hcsp.Sequence(
            hcsp.OutputChannel('ch_clock', AConst(1)),
            hcsp.OutputChannel('ch_clock', AConst(0)))))



from ss2hcsp.sl.sl_block import SL_Block
from ss2hcsp.hcsp.expr import AVar, true_expr, OpExpr, RelExpr, AConst
from ss2hcsp.hcsp import hcsp as hp


class Product(SL_Block):
    """Multiply (or divide) a list of dest lines."""
    def __init__(self, name, dest_spec, st=-1):
        """dest_spec is a list of either '*' or '/'."""
        super(Product, self).__init__("product", name, 1, len(dest_spec), st)

        # dest_spec is a list of either '*' or '/'
        assert all(s == '*' or s == '/' for s in dest_spec)
        assert len(dest_spec) >= 2
        self.dest_spec = dest_spec  # string

    def get_expr(self):
        """Compute the assignment corresponding to a product block."""
        in_vars = [AVar(line.name) for line in self.dest_lines]
        if self.dest_spec[0] == '*':
            expr = in_vars[0]
        else:
            expr = OpExpr("/", AConst(1), in_vars[0])
        for op, in_var in zip(self.dest_spec[1:], in_vars[1:]):
            if op == '*':
                expr = OpExpr("*", expr, in_var)
            else:
                expr = OpExpr("/", expr, in_var)
        return expr

    def __str__(self):
        expr = self.get_expr()
        out_var = self.src_lines[0][0].name
        return "%s: %s = %s  (st = %s)" % (self.name, out_var, str(expr), str(self.st))

    def __repr__(self):
        return "Product(%s, %s, %s, in = %s, out = %s)" % \
            (self.name, self.dest_spec, self.st, self.dest_lines, self.src_lines)

    def get_output_hp(self):
        expr = self.get_expr()
        out_var = self.src_lines[0][0].name
        return hp.Assign(out_var, expr)

    def get_var_subst(self):
        expr = self.get_expr()
        out_var = self.src_lines[0][0].name
        return {out_var: expr}

    def get_var_map(self):
        expr = self.get_expr()
        out_var = self.src_lines[0][0].name
        return {out_var: [(true_expr, expr)]}
